open Common

type t = 
| EVar of string * int
| ENext of string * int
| EExists of t list * t
| EForall of t list * t
| ENot of t
| EAnd of t * t
| EOr of t * t
| EList of t list
| True 
| False

let rec to_string = function 
  | EVar (s,i) -> Printf.sprintf "%s.(%d)" s i
  | ENext (s,i) -> Printf.sprintf "next %s.(%d)" s i
  | EExists (list,e) -> Printf.sprintf "(exists [%s]. %s)" (to_string (EList list)) (to_string e)
  | EForall (list,e) -> Printf.sprintf "(forall [%s]. %s)" (to_string (EList list)) (to_string e)
  | ENot a -> Printf.sprintf "not %s" (to_string a)
  | EAnd (a,b) -> Printf.sprintf "(%s & %s)" (to_string a) (to_string b)
  | EOr (a,b) -> Printf.sprintf "(%s || %s)" (to_string a) (to_string b)
  | EList l -> List.fold_left (fun a b -> a ^ (to_string b)^"; ") "" l
  | True -> "True"
  | False -> "False"
    
    
let var s i = EVar (s,i)
let next_var s i = ENext (s,i)
  
exception AlreadyNext of t
let rec next e = match e with
    | EVar (s,i) -> ENext (s,i)
    | ENext(_,_) -> raise (AlreadyNext e)
    | EExists (el,e) -> EExists (List.map next el, next e)
    | EForall (el,e) -> EForall (List.map next el, next e)
    | ENot x -> ENot (next x)
    | EAnd (a,b) -> EAnd (next a,next b)
    | EOr (a,b) -> EOr (next a,next b)
    | EList (el) -> EList (List.map next el)
    | True -> True
    | False -> False


let forall vl e = match e with 
    | True | False -> e
    | _ -> EForall (vl,e)
let exists vl e = match e with 
    | True | False -> e
    | _ -> EExists (vl,e)

let of_list el = EList el
    

let disjunction a b = match a , b with
    | True, x | x , True -> True
    | False, x | x, False -> x
    | a, b -> if a = b then a else if a = ENot b or ENot a = b then True else EOr (a,b)
let conjunction a b = match a , b with
    | True, x | x , True -> x
    | False, x | x, False -> False
    | a, b -> if a = b then a else if a = ENot b or ENot a = b then False else EAnd (a,b)
	
let negation a = match a with
  | True -> False | False -> True | ENot a -> a | x -> ENot x
    
let xor a b = disjunction (conjunction a (negation b)) (conjunction (negation a) b)
  
let equality a b = disjunction (conjunction a b) (conjunction (negation a) (negation b))
  
let implication a b = disjunction b (negation a)
  
let for_each list f =
  let treat_one (start,last) =
    of_list (iter start last (fun i c -> (f i) :: c) [])
  in of_list (List.map treat_one list)
  


let of_bdd bdd symbols = 
  let insert bdd expr = 
    let rec aux accu = function
      | [] -> List.rev_append accu [(bdd,expr)]
	| (a,f) :: r -> 
	  let c = Cudd.compare a bdd in
	  if c < 0 then aux ((a,f)::accu) r
	  else if c > 0 then List.rev_append accu ((bdd,expr) :: (a,f) :: r)
	  else List.rev_append accu ((bdd, disjunction expr f)::r)
      in aux [] 
    in
    
    let aux bdd_expr_list (name,opt) = 
      let var = AigerBdd.Variable.find (name,opt) in
      List.fold_left 
	(fun accu (bdd,expr) ->
	  let var_bdd = AigerBdd.Variable.to_bdd var in
	  let t = Cudd.bddRestrict bdd var_bdd in
	  let expr_t = conjunction expr (EVar (name,opt)) in
          let e = Cudd.bddRestrict bdd (Cudd.bddNot var_bdd) in
	  let expr_e = conjunction expr (ENot (EVar (name,opt))) in
	  
	  if Cudd.compare e t = 0
	  then insert e expr accu 
	  else
	    let accu = insert t expr_t accu in
	    let accu = insert e expr_e accu in
	    accu
	) [] bdd_expr_list 
    in 
    
    let res = List.fold_left aux [bdd,True] symbols in
    match List.filter (fun (bdd,expr) -> Cudd.equal bdd (Cudd.bddTrue())) res with 
    | [] -> False
    | (bdd,expr) :: [] -> expr
    | _ -> failwith "in [of_bdd]: missing variables"



