open Attractor

type value = Winning | CoopWinning | Losing | NotWinning | NotLosing

(* Values have the following meaning:
 * 1 -> there is a winning strategy
 * 0 -> there is no winning strategy but we can win if other players cooperate
 * -1 -> it is impossible to win
 *)

let winning_with_restriction restriction p controllables uncontrollables ?(weak=false) unsafe = 
  let states = Cudd.bddNot (trap_with_restriction (Cudd.bddTrue(),fst restriction,snd restriction) p controllables uncontrollables ~weak unsafe) in
  let transition = Cudd.bddVectorCompose (Cudd.bddSwapVariables states (AigerBdd.array_variables p) (AigerBdd.array_next_variables p)) (AigerBdd.composition_vector p) in
  let controllable_cube = Cudd.make_cube (List.map Aiger.lit2int controllables) in
  if weak
  then
    let restricted = Cudd.bddAnd transition (snd restriction) in
    let restricted = Cudd.bddRestrict restricted (Cudd.bddOr (fst restriction) restricted) in
    let uncontrollable_cube = Cudd.make_cube (List.map Aiger.lit2int uncontrollables) in
    let univ_abstract = Cudd.bddUnivAbstract restricted uncontrollable_cube in
    (states,Cudd.bddExistAbstract univ_abstract controllable_cube)
  else
    (states,Cudd.bddExistAbstract (Cudd.bddAnd transition (fst restriction)) controllable_cube)

let winning x = winning_with_restriction (Cudd.bddTrue(),Cudd.bddTrue()) x

let losing_with_restriction restriction p controllables uncontrollables unsafe = 
  let states = trap_with_restriction (Cudd.bddTrue(),fst restriction,snd restriction) p (List.rev_append controllables uncontrollables) [] unsafe in
  let transition = Cudd.bddVectorCompose (Cudd.bddSwapVariables states (AigerBdd.array_variables p) (AigerBdd.array_next_variables p)) (AigerBdd.composition_vector p) in
  let controllable_cube = Cudd.make_cube (List.map Aiger.lit2int controllables) in
  (states,Cudd.bddUnivAbstract (Cudd.bddOr transition (Cudd.bddNot (fst restriction))) controllable_cube)

let losing x = losing_with_restriction (Cudd.bddTrue(),Cudd.bddTrue()) x

let value_with_restriction restriction p controllables uncontrollables ?(weak=false) unsafe = 
  let ws,wp = winning_with_restriction restriction p controllables uncontrollables ~weak unsafe in
  let ls,lp = losing_with_restriction restriction p controllables uncontrollables unsafe in
  let cs,cp = Cudd.bddAnd (Cudd.bddNot ws) (Cudd.bddNot ls),
    Cudd.bddAnd (Cudd.bddNot wp) (Cudd.bddNot lp) in
  function 
  | Winning -> ws,wp
  | Losing -> ls,lp
  | CoopWinning -> cs,cp
  | NotWinning -> Cudd.bddNot ws,Cudd.bddNot wp
  | NotLosing -> Cudd.bddNot ls,Cudd.bddNot lp
     
let value x = value_with_restriction (Cudd.bddTrue (),Cudd.bddTrue()) x

let strategies_with_restriction restriction p v =
  Printf.printf "warning restriction not used\n";
  let successorNotWinning = Cudd.bddVectorCompose (Cudd.bddSwapVariables (fst (v NotWinning)) (AigerBdd.array_variables p) (AigerBdd.array_next_variables p)) (AigerBdd.composition_vector p) in
  let successorLosing = Cudd.bddVectorCompose (Cudd.bddSwapVariables (fst (v Losing)) (AigerBdd.array_variables p) (AigerBdd.array_next_variables p)) (AigerBdd.composition_vector p) in
  Cudd.bddAnd 
    (Cudd.bddNot (Cudd.bddAnd (snd (v Winning)) successorNotWinning))
    (Cudd.bddNot (Cudd.bddAnd (snd (v NotLosing)) successorLosing))

let strategies x = strategies_with_restriction (Cudd.bddTrue()) x

let strongly_winning_strategy p winning = failwith "unimplemented"

let correct_strategy p strat controllables uncontrollables =
  let controllable_cube = Cudd.make_cube (List.map Aiger.lit2int controllables) in
  let uncontrollable_cube = Cudd.make_cube (List.map Aiger.lit2int uncontrollables) in
  let variables_cube = Cudd.make_cube (AigerBdd.variables p) in
  let exist_quantified = Cudd.bddExistAbstract strat controllable_cube in
  let univ_quantified = Cudd.bddUnivAbstract exist_quantified uncontrollable_cube in
  Cudd.bddUnivAbstract univ_quantified variables_cube

let check_strategies p strat controllables uncontrollables u =
  let controllable_cube = Cudd.make_cube (List.map Aiger.lit2int controllables) in
  let uncontrollable_cube = Cudd.make_cube (List.map Aiger.lit2int uncontrollables) in
  let unsafe_pred u =
    let renamed = Cudd.bddSwapVariables u (AigerBdd.array_variables p) (AigerBdd.array_next_variables p) in
    let substitued = Cudd.bddVectorCompose renamed (AigerBdd.composition_vector p) in
    let apply_strat = Cudd.bddAnd substitued strat in
    let exist_quantified = Cudd.bddExistAbstract apply_strat controllable_cube in
    let exist_quantified1 = Cudd.bddExistAbstract exist_quantified uncontrollable_cube in
    exist_quantified1
  in
  let rec trap u = 
    let x = unsafe_pred u in
    let res = Cudd.bddOr u x in
    if Cudd.equal res u then res else trap res
  in trap u 

let assume_admissible game player1_inputs player2_inputs fail1 fail2 = 
  let v1 = value game player1_inputs player2_inputs fail1 in
  let adm1 = strategies game v1 in
  let v2 = value game player2_inputs player1_inputs ~weak:true fail2 in
  let adm2 = strategies game v2 in
  let w1 = Cudd.bddNot (Attractor.trap_with_restriction (Cudd.bddTrue(),adm1,adm2) game player1_inputs player2_inputs fail1) in
  let w2 = Cudd.bddNot (Attractor.trap_with_restriction (Cudd.bddTrue(),adm2,adm1) game player2_inputs player1_inputs ~weak:true fail2) in
  Cudd.bddAnd w1 w2


let test aiger = 
  AigerBdd.init aiger;
  (* let unsafe = List.fold_left Cudd.bddAnd (Cudd.bddTrue()) (List.map (fun x -> Cudd.ithVar (Aiger.lit2int x)) aiger.Aiger.outputs) in *)

  
  let fails1 = List.filter (fun l -> Aiger.lit2string aiger l = "player_1_err") aiger.Aiger.outputs
  in
  let fail1 = List.fold_left Cudd.bddAnd (Cudd.bddTrue()) (List.map AigerBdd.bdd_of_lit fails1) in 

  let fails2 = List.filter (fun l -> Aiger.lit2string aiger l = "player_2_err") aiger.Aiger.outputs
  in
  let fail2 = List.fold_left Cudd.bddAnd (Cudd.bddTrue()) (List.map AigerBdd.bdd_of_lit fails2) in 

  let inputs1 = List.filter 
    (fun l -> Common.starts_with "player_1_" (fst (Aiger.lit2symbol aiger l))
    ) aiger.Aiger.inputs 
  in

  let inputs2 = List.filter 
    (fun l -> Common.starts_with "player_2_" (fst (Aiger.lit2symbol aiger l))
    ) aiger.Aiger.inputs 
  in

  let initial = AigerBdd.initial_state aiger in
  let p = AigerBdd.of_aiger aiger in
  let win = assume_admissible p inputs1 inputs2 fail1 fail2 in
  Cudd.dumpDot "assume_admissible.dot" win;
  let initial_winning = Cudd.bddRestrict win initial in
  
  if Cudd.value initial_winning = 1 
  then print_endline "assume admissible has a solution"
  else if Cudd.value initial_winning = 0
  then print_endline "no solution"
  else (Cudd.dumpDot ("problem_report.out") initial_winning;  print_endline "problem...")


(*

  let strat = 
    (*if not comp_synt else
	 let strat_comp = compositional_synthesis p contr uncontr unsafe in
	 strat_comp    then *)
    let v = value p contr uncontr unsafe in
    Cudd.dumpDot ("value1.dot") (fst (v Winning));
    Cudd.dumpDot ("value0.dot") (fst (v CoopWinning));
    Cudd.dumpDot ("value-1.dot") (fst (v Losing));
    let strat = strategies p v in
    strat
  in

  Cudd.dumpDot ("strategy.dot") strat;


  print_endline "checking correctness";
  let correct = correct_strategy p strat contr uncontr in
  if Cudd.value correct = 1 
  then print_endline "correct strategy"
  else if Cudd.value correct = 0
  then print_endline "uncorrect strategy"
  else print_endline "problem checking correctness...";
  ();
*)

  
  (*
  let rec loop strat bound =
    print_endline "checking";
    let trap_set = check_strategies p strat contr uncontr unsafe in
    let initial_losing = Cudd.bddRestrict trap_set initial in
  
    if Cudd.value initial_losing = 1 
    then 
    (
      print_endline "unsafe strategy";
      if bound > 0 
      then
	(
	  print_endline "trying to refine it...";
	  let refine = refinement strat p contr uncontr unsafe in
	  if Cudd.equal refine strat 
	  then loop refine 0
	  else loop refine (bound - 1)
	)
      else
	(print_endline "trying to correct it...";
	 let trap_set = Attractor.trap_with_restriction (strat,Cudd.bddTrue()) p contr uncontr unsafe in
	 print_endline "attractor computed";
	 let initial_losing = Cudd.bddRestrict trap_set initial in
	 if Cudd.value initial_losing = 1 
	 then print_endline "unrealizable"
	 else if Cudd.value initial_losing = 0
	 then print_endline "realizable";)
    )
    else if Cudd.value initial_losing = 0
    then print_endline "safe strategy"
    else (Cudd.dumpDot ("problem_report.dot") initial_losing;  print_endline "problem...")
  in loop strat 5*)


let main =
  if Filename.check_suffix Sys.argv.(0) "admissibility" 
  then
    if Array.length Sys.argv < 2 
    then print_endline "usage : admissibility <file>"
    else 
      let aiger = Aiger.read_from_file Sys.argv.(1) in
      test aiger

