(** Values have the following meaning:
    * Winning -> there is a winning strategy
    * CoopWinning -> there is no winning strategy but we can win if other players cooperate
    * Losing -> it is impossible to win
    * NotWinning -> there is no winning strategy (union of the two preceding cases)
    * NotLosing -> it is possible to win (union of the two first cases)
 *)
type value = Winning | CoopWinning | Losing | NotWinning | NotLosing

(** [value aiger controllables uncontrollables failure i] 
    Compute the set of states which have value [i].
    It returns two BDDs, the first represents the configurations of the latches that have value [i], the other represents the pairs of configuration and action of the environment that have value [i].
    To avoid computing the same sets several time, you should first compute [value aiger controllables uncontrollables failure] then call this on the [i]'s you need.
    If weak is set to true then the controller as to provide its action before the environment.
*)
val value : AigerBdd.t -> Aiger.lit list -> Aiger.lit list -> ?weak:bool -> Cudd.bdd -> value -> (Cudd.bdd * Cudd.bdd)

(* the restriction should be first on controllable actions and then uncontrollables ? *)
val value_with_restriction : (Cudd.bdd * Cudd.bdd) -> AigerBdd.t -> Aiger.lit list -> Aiger.lit list -> ?weak:bool -> Cudd.bdd -> value -> (Cudd.bdd * Cudd.bdd)

(** Makes a safe version of the given operator: never increase the size of the BDD. *)
val safe : (Cudd.bdd -> Cudd.bdd -> Cudd.bdd) -> Cudd.bdd -> Cudd.bdd -> Cudd.bdd


val strategies : AigerBdd.t -> (value -> (Cudd.bdd * Cudd.bdd)) -> Cudd.bdd
val strategies_with_restriction : (Cudd.bdd * Cudd.bdd) -> AigerBdd.t -> (value -> (Cudd.bdd * Cudd.bdd)) -> Cudd.bdd


(** The second argument is a BDD representing the set of winning states *)
val strongly_winning_strategy : AigerBdd.t -> Cudd.bdd -> Cudd.bdd

(** The arguments are: the game, list of inputs controlled by player 1, list of inputs controlled by player 2, states to be avoided by player 1, states to be avoided by player 2. 
We should assume that player 2 choses its inputs before player 1. *)
val assume_admissible : AigerBdd.t -> Aiger.lit list -> Aiger.lit list -> Cudd.bdd -> Cudd.bdd -> Cudd.bdd

(** Given a strategy, computes the set of unsafe states.
    The function checks that the strategy is safe but not that it is a correct strategy (ie: at least one action is possible in each state)
*)

val check_strategies : AigerBdd.t -> Cudd.bdd -> Aiger.lit list -> Aiger.lit list -> Cudd.bdd -> Cudd.bdd

(** The function checks that the strategy is a correct strategy (ie: at least one action is possible in each state)*)
val correct_strategy : AigerBdd.t -> Cudd.bdd -> Aiger.lit list -> Aiger.lit list -> Cudd.bdd
