(** This module defines usefull function for going from the aiger representation 
 * to the BDD representation and reciproqually *)

let init aiger =
  try Cudd.init (aiger.Aiger.num_inputs+2*aiger.Aiger.num_latches+2)
  with Failure x -> Printf.eprintf "warning: %s\n" x

let bdd_of_gate lit = 
  if lit = Aiger.aiger_true then Cudd.bddTrue() 
  else if lit = Aiger.aiger_false then Cudd.bddFalse ()
  else 
    if Aiger.sign lit
    then Cudd.bddNot (Cudd.ithVar (Aiger.lit2int (Aiger.strip lit)))
    else Cudd.ithVar (Aiger.lit2int lit)


type variable = int

let bdd_of_variable var = 
  if var mod 2 = 1
  then failwith (Printf.sprintf "variable should not be a next : %d" var)
  else Cudd.ithVar var

let bdd_of_not_gate lit = Cudd.bddNot (bdd_of_gate lit)
let bdd_of_not_variable v = Cudd.bddNot (bdd_of_variable v)

let bdd_of_next_variable var = 
  if var mod 2 = 1
  then failwith (Printf.sprintf "variable should not be a next : %d" var)
  else Cudd.ithVar (1 + var)
    
let next var = 
  if var mod 2 = 1
  then failwith "variables is already next"
  else var + 1

let make_cube lit_list = Cudd.make_cube (List.map Aiger.lit2int lit_list)

type valuation = bool Aiger.LitMap.t




exception UndeclaredLit of Aiger.lit

(* We should normalize the cache: ie no negated nodes *)
let add_bdd_to_aiger aiger bdd = 
  let cache = Hashtbl.create 1000 in

  let rec aux bdd aig = 
    try (aig, Hashtbl.find cache (Cudd.id bdd))
    with Not_found ->
      if Cudd.isConstant bdd
      then
	if Cudd.value bdd = 1
	then (aig,Aiger.aiger_true)
	else (aig,Aiger.aiger_false)
      else
	let lit = Aiger.int2lit (Cudd.nodeReadIndex bdd) in
	let then_child = Cudd.t bdd in
	let else_child = Cudd.e bdd in
	let (aig,then_lit) = aux then_child aig in
	let (aig,else_lit) = aux else_child aig in

	if Aiger.lit2int lit <= 1 then Common.debug "lit <= 1\n"; 
	
	let aig, lit1 =
	  if then_lit = Aiger.aiger_false 
	  then aig, Aiger.aiger_false
	  else if then_lit = Aiger.aiger_true
	  then aig, lit
	  else
	    let aig,var1 = Aiger.new_var aig in 
	    let lit1 = Aiger.var2lit var1 in
	    let aig = Aiger.add_and aig lit1 lit then_lit in
	    aig, lit1
	in

	let aig, lit2 =
	  if else_lit = Aiger.aiger_false 
	  then aig, Aiger.aiger_false
	  else if else_lit = Aiger.aiger_true
	  then aig, Aiger.aiger_not lit
	  else
	    let aig,var2 = Aiger.new_var aig in 
	    let lit2 = Aiger.var2lit var2 in
	    let aig = Aiger.add_and aig lit2 (Aiger.aiger_not lit) else_lit in
	    aig, lit2
	in

	let aig,var3 = Aiger.new_var aig in
	let lit3 = Aiger.var2lit var3 in
	let aig = Aiger.add_and aig lit3 (Aiger.aiger_not lit1) (Aiger.aiger_not lit2) in
	
	let res = 
	  if Cudd.isComplement bdd then lit3 else Aiger.aiger_not lit3
	in
	Hashtbl.add cache (Cudd.id bdd) res;

	(aig,res)

  in 
(*  try *)
    aux bdd aiger 
(*  with Aiger.BadAnd(x,y,z) -> 
    Printf.printf "bad and %d %d %d\n" (Aiger.lit2int x) (Aiger.lit2int y) (Aiger.lit2int z);
    raise (UndeclaredLit(x))
*)


let bdd_to_aiger inputs bdd =
  ()

let variable_of_output o = 
  if Aiger.lit2int o < 2
  then (Common.debug "warning : there is a constant output\n"; 
	Aiger.lit2int (Aiger.strip o))
  else if Aiger.sign o 
  then (Common.debug "warning : there is a negated output\n"; 
	Aiger.lit2int (Aiger.strip o))
  else Aiger.lit2int o

let variable_of_latch2 l = 
  Aiger.lit2int (fst l)

let variable_of_latch l = 
  Aiger.lit2int l

let variable_of_input i = 
  Aiger.lit2int i


let variables_aiger aiger = 
  List.rev_append 
    (List.map variable_of_output aiger.Aiger.outputs) 
    (List.map variable_of_latch2 aiger.Aiger.latches)

let bdd_of_valuation valuation =
    Aiger.LitMap.fold 
    (fun lit bool accu -> 
      if bool
      then Cudd.bddAnd accu (bdd_of_variable (variable_of_latch lit))
      else Cudd.bddAnd accu (bdd_of_not_variable (variable_of_latch lit))
    ) valuation (Cudd.bddTrue ())

let bdd_of_inputs input_valuation =
  Aiger.LitMap.fold
    (fun lit b accu -> 
     Cudd.bddAnd accu (if b then bdd_of_variable (variable_of_input lit) else bdd_of_not_variable (variable_of_input lit))
    ) input_valuation (Cudd.bddTrue ())

let bdd_to_valuations bdd variables = 
  let rec aux bdd valuations = function 
    | [] -> valuations
    | var :: s ->
      let restricted_true = Cudd.bddRestrict bdd (bdd_of_variable (variable_of_latch var)) in
      let restricted_false = Cudd.bddRestrict bdd (bdd_of_not_variable (variable_of_latch var)) in
      if Cudd.equal restricted_true (Cudd.bddFalse ())
      then 
	aux restricted_false (List.map (Aiger.LitMap.add var false) valuations) s
      else
	if Cudd.equal restricted_false (Cudd.bddFalse ())
	then 
	  aux restricted_true (List.map (Aiger.LitMap.add var true) valuations) s
	else
	  let a_true = aux restricted_true (List.map (Aiger.LitMap.add var true) valuations) s in
	  let a_false = aux restricted_false (List.map (Aiger.LitMap.add var false) valuations) s
	  in List.rev_append a_true a_false
  in  
  if Cudd.equal bdd (Cudd.bddFalse())
  then []
  else aux bdd [Aiger.LitMap.empty] variables


exception Undefined of string 

let extract_valuation aiger valuation names =
  List.map
    (fun name -> 
      try 
	let size = Aiger.size_symbol aiger name in
	if size = 1 
	then 
	  if Aiger.LitMap.find (Aiger.symbol2lit aiger (name,None)) valuation
	  then name,1 else name,0
	else
	  let sum = ref 0 in
	  for i = 0 to size - 1 do
	    if Aiger.LitMap.find (Aiger.symbol2lit aiger (name,Some i)) valuation 
	    then sum := !sum + (1 lsl i)
	  done;
	  name, !sum
      with Not_found -> raise (Undefined name)
    ) names





(** compute updates in a table, all variables are even *)
let compute_updates aiger =
  let gate_bdd = Hashtbl.create aiger.Aiger.maxvar in
  let updates = Hashtbl.create aiger.num_latches in
  Hashtbl.add gate_bdd Aiger.aiger_false (Cudd.bddFalse ());
  Hashtbl.add gate_bdd Aiger.aiger_true (Cudd.bddTrue ());

  List.iter 
    (fun inp -> 
      let i = Aiger.lit2int inp in
      Hashtbl.add gate_bdd inp (Cudd.ithVar i);
      Hashtbl.add gate_bdd (Aiger.aiger_not inp) (Cudd.bddNot (Cudd.ithVar i))
    ) aiger.inputs;

  List.iter 
    (fun (l,_) -> 
      Hashtbl.add gate_bdd l (bdd_of_gate l);
      Hashtbl.add gate_bdd (Aiger.aiger_not l) (bdd_of_not_gate l)
    ) aiger.latches;

  List.iter 
    (fun (g,l,r) -> 
      try
	let gl = Hashtbl.find gate_bdd l in
	let gr = Hashtbl.find gate_bdd r in
	let b = Cudd.bddAnd gl gr in
	Hashtbl.add gate_bdd g b;
	Hashtbl.add gate_bdd (Aiger.aiger_not g) (Cudd.bddNot b)
      with Not_found -> 
	Printf.eprintf "gate %d or %d not found\n" (Aiger.lit2int l) (Aiger.lit2int r);
	raise Not_found
    ) aiger.ands;

  List.iter 
    (fun (l,n) -> 
      try Hashtbl.add updates (variable_of_latch l) (Hashtbl.find gate_bdd n)
      with Not_found -> 
	Printf.eprintf "gate %d not found\n" (Aiger.lit2int n);
	raise Not_found 
    ) aiger.latches;

  List.iter 
    (fun o -> 
      try Hashtbl.add updates (variable_of_output o) (Hashtbl.find gate_bdd o)
      with Not_found  -> 
	Printf.eprintf "gate %d not found\n" (Aiger.lit2int o);
	raise Not_found
    ) aiger.outputs;

  updates



type t = 
  {
    updates:(variable , Cudd.bdd) Hashtbl.t;
    variables: int list;
    next_variables: int list;
    array_variables: int array;
    array_next_variables: int array;
    composition_vector: Cudd.bdd array;
  }

let updates p = p.updates
let variables p = p.variables
let next_variables p = p.next_variables
let array_variables p = p.array_variables
let array_next_variables p = p.array_next_variables
let composition_vector p = p.composition_vector

let of_aiger aiger = 
  Common.debug "compute updates";
  let updates = compute_updates aiger in
  Common.debug "variables";
  let variables = variables_aiger aiger in
  Common.debug "next variables";
  let next_variables = List.map ((+) 1) variables in

  if !Common.display_debug
  then (Printf.eprintf "maxvar=%d\n" aiger.Aiger.maxvar; prerr_endline "composition vector");
  let composition_vector = Array.init (aiger.Aiger.maxvar * 2 + 2) (* should it really be this value ? *)
    (fun i -> 
      try Hashtbl.find updates (i-1)
      with Not_found -> Cudd.bddTrue())
  in

  Common.debug "composition vector finished";
  let array_variables = Array.of_list variables in
  let array_next_variables = Array.of_list next_variables in
  { updates = updates; variables=variables; next_variables = next_variables;
    array_variables=array_variables; array_next_variables = array_next_variables;
    composition_vector=composition_vector;}

let rename_configuration bdd variables next_variables =
  Cudd.bddSwapVariables bdd variables next_variables

let rec fixpoint f x =
  let y = f x in if Cudd.equal y x then x else fixpoint f y



let initial_state aiger = 
  let init = 
    List.fold_left Cudd.bddAnd (Cudd.bddTrue()) 
      (List.map bdd_of_not_variable 
	 (List.map variable_of_latch2 aiger.Aiger.latches)) in
  List.fold_left Cudd.bddAnd init 
    (List.map bdd_of_not_variable
       (List.map variable_of_output aiger.Aiger.outputs)) 


let valuation_of_list list = 
  List.fold_left (fun accu (x,b) -> Aiger.LitMap.add x b accu) Aiger.LitMap.empty list


let successor_states aiger state inputs = 
  let update_table = updates aiger in
  let state_with_inputs = Cudd.bddAnd (bdd_of_inputs inputs) (bdd_of_valuation state) 
  in
  let new_state =
    Hashtbl.fold
      (fun k bdd accu -> 
	try Cudd.bddAnd accu
	      (Cudd.bddOr
		 (Cudd.bddAnd bdd (bdd_of_next_variable k))
		 (Cudd.bddAnd (Cudd.bddNot bdd) (Cudd.bddNot (bdd_of_next_variable k))))

	with Not_found -> Printf.eprintf "literal %d not found\n" k;
	  raise Not_found
      ) update_table state_with_inputs
  in

  let var_list = Aiger.LitMap.fold (fun x b accu -> x :: accu) state [] in
  let input_list = Aiger.LitMap.fold (fun x b accu -> x :: accu) inputs [] in
  let cube = make_cube (List.rev_append var_list input_list) in
(*
    Cudd.make_cube (List.map (fun (x,_) -> Aiger.lit2int x) *)

  let new_state_exist = Cudd.bddExistAbstract new_state cube in
  let variables = Hashtbl.fold (fun k _ accu -> (*Aiger.lit2int*) k :: accu) update_table [] in
  let next_variables = List.map next variables in
  let array_variables = Array.of_list variables in
  let array_next_variables = Array.of_list next_variables in
  let renamed = 
    rename_configuration new_state_exist array_next_variables array_variables
  in 
  bdd_to_valuations renamed (List.map Aiger.int2lit variables)
