(** This module defines usefull function for going from the aiger representation 
 * to the BDD representation and reciproqually *)

(** Initialize CUDD. 
    The number of variables to be used is inferred from the 
    circuit that is provided. *)
val init : Aiger.t -> unit

val bdd_of_gate : Aiger.lit -> Cudd.bdd
val bdd_of_not_gate : Aiger.lit -> Cudd.bdd

val make_cube : Aiger.lit list -> Cudd.cube

type variable 
val variable_of_output : Aiger.lit -> variable
val variable_of_latch2 : (Aiger.lit*Aiger.lit) -> variable
val variable_of_latch : Aiger.lit -> variable
val variable_of_input : Aiger.lit -> variable
val bdd_of_variable : variable -> Cudd.bdd
val bdd_of_not_variable : variable -> Cudd.bdd
val bdd_of_next_variable : variable -> Cudd.bdd
val next : variable -> variable

(** Raised by [bdd_to_aiger] if one of the literals in the BDD is not already in the given aiger *)
exception UndeclaredLit of Aiger.lit

(** Add to the aiger file, gate to compute the result of the 
    evaluation of the BDD. The result contains the new aiger
    file and the literal that points to the result. 
    All the variables of the BDD should correspond to literals of the given aiger file,
    otherwise a [UndeclaredLit] exception will be raised.
*)
val add_bdd_to_aiger : Aiger.t -> Cudd.bdd -> (Aiger.t * Aiger.lit)

(** Takes as argument the list of inputs of the circuit to produce *)
val bdd_to_aiger : Aiger.LitMap.t -> Cudd.bdd -> Aiger.t


(** Valuation of registers of the aiger representation. *)
type valuation = bool Aiger.LitMap.t

val valuation_of_list : (Aiger.lit * bool) list -> valuation

(** Convert the valuation to a BDD representation *)
val bdd_of_valuation : valuation -> Cudd.bdd

val bdd_to_valuations : Cudd.bdd -> Aiger.lit list -> valuation list

val extract_valuation : Aiger.t -> valuation -> string list -> (string * int) list



type t 

val of_aiger : Aiger.t -> t

(** Updates of the latches are stocked as BDDs. 
    The returned table contain such a BDD for each latch and each output *)
val updates : t -> (variable , Cudd.bdd) Hashtbl.t
val variables : t -> variable list
val next_variables : t -> variable list
val array_variables : t -> int array
val array_next_variables : t -> int array
val composition_vector : t -> Cudd.bdd array

(** Fixpoint of an operation on BDD's *)
val fixpoint : (Cudd.bdd -> Cudd.bdd) -> Cudd.bdd -> Cudd.bdd

(** Compute a BDD representing the initial state, in which all latches and output are set to false. *)
val initial_state : Aiger.t -> Cudd.bdd

(** [successor_state updates state inputs] gives one possible next state *)
val successor_states : t -> valuation ->  valuation -> valuation list

val rename_configuration : Cudd.bdd -> variable array -> variable array -> Cudd.bdd
