open AigerBdd

let controllable_name name = 
     if String.length name > 13 then
       String.sub name 0 13 = "controllable_" 
     else false

let controllables aiger = 
  List.fold_left
    (fun (c,u) lit -> 
     let name,_ = Aiger.lit2symbol aiger lit in
     if controllable_name name then (lit :: c, u) 
     else (c, lit :: u)
    ) ([],[]) aiger.Aiger.inputs 


let of_aiger aiger = 
  let c,u = controllables aiger in
  (aiger, c, u)
    
let read_from_file file = 
  let aiger = Aiger.read_from_file file in
  of_aiger aiger

type t = { latch_configuration : Cudd.bdd; latch_input_configuration : Cudd.bdd }
  
let make x y = {latch_configuration=x; latch_input_configuration= y}
let of_bdds x y = make x y

let latch_configuration x = x.latch_configuration 
let latch_input_configuration x = x.latch_input_configuration 
let union x y = make (Cudd.bddOr (latch_configuration x) (latch_configuration y)) (Cudd.bddOr (latch_input_configuration x) (latch_input_configuration y)) 


let equal x y =  
  (Cudd.equal (latch_configuration x) (latch_configuration y))
  && (Cudd.equal (latch_input_configuration x) (latch_input_configuration y))

let ff () = make (Cudd.bddFalse()) (Cudd.bddFalse())
let tt () = make (Cudd.bddTrue()) (Cudd.bddTrue())
let intersection x y = make (Cudd.bddAnd (latch_configuration x) (latch_configuration y)) (Cudd.bddAnd (latch_input_configuration x) (latch_input_configuration y)) 

let intersection_bdd x bdd = make (Cudd.bddAnd (latch_configuration x) bdd) (Cudd.bddAnd (latch_input_configuration x) bdd)

let negation x = make (Cudd.bddNot (latch_configuration x)) (Cudd.bddNot (latch_input_configuration x))
let diff x y = intersection x (negation y)


let rec fixpoint f x = 
  let next = union (f x) x in
  if equal next x then x else fixpoint f next



let rename_substitute_restrict unsafe aiger =
  let renamed = Cudd.bddSwapVariables (latch_configuration unsafe) (AigerBdd.array_variables aiger) (AigerBdd.array_next_variables aiger) in
  let substitued = Cudd.bddVectorCompose renamed (AigerBdd.composition_vector aiger) in
  Cudd.bddOr substitued (latch_input_configuration unsafe)

let uncontrollable_predecessors_with_restriction unsafe aiger controllable_cube uncontrollable_cube weak restriction =
  let restricted = rename_substitute_restrict unsafe aiger in
  if weak 
  then 
    let exist_quantified = Cudd.bddExistAbstract (Cudd.bddAnd (latch_configuration restriction) restricted) uncontrollable_cube in
    let univ_quantified = Cudd.bddUnivAbstract (Cudd.bddOr exist_quantified (Cudd.bddNot (latch_configuration restriction))) controllable_cube in
    make univ_quantified exist_quantified
  else
    let univ_quantified = Cudd.bddUnivAbstract (Cudd.bddOr restricted (Cudd.bddNot (latch_configuration restriction))) controllable_cube in
    let exist_quantified = Cudd.bddExistAbstract (Cudd.bddAnd (latch_input_configuration restriction) univ_quantified) uncontrollable_cube in
    make exist_quantified univ_quantified


let trap_with_restriction aiger controllables uncontrollables ?(weak=false) unsafe restriction =
  let controllable_cube = AigerBdd.make_cube controllables in
  let uncontrollable_cube = AigerBdd.make_cube uncontrollables in

  let aux u = 
    uncontrollable_predecessors_with_restriction u aiger controllable_cube uncontrollable_cube weak restriction
  in 
  fixpoint aux unsafe


let trap aiger contr uncontr ?(weak=false) unsafe =
  trap_with_restriction aiger contr uncontr ~weak (make unsafe (Cudd.bddFalse())) (tt())

let attractor_with_restriction aiger controllables uncontrollables ?(weak=false) safe restriction =
  trap_with_restriction aiger uncontrollables controllables ~weak:(not weak) safe restriction

let attractor aiger contr uncontr ?(weak=false) safe = 
  attractor_with_restriction aiger contr uncontr ~weak (make safe (Cudd.bddFalse())) (tt())

let strategy game region =
  Cudd.bddVectorCompose (Cudd.bddSwapVariables
			   (latch_configuration region)
			   (AigerBdd.array_variables game) 
			   (AigerBdd.array_next_variables game))
			(AigerBdd.composition_vector game)

let strategy_to_aiger aiger strategy controllables uncontrollables= 
  let cube = AigerBdd.make_cube controllables in
  let circuit = 
    List.fold_left
      (fun c inp -> 
	Aiger.add_input c inp (Aiger.lit2symbol aiger inp)
      ) Aiger.empty uncontrollables
  in
  let circuit = 
    List.fold_left
      (fun c (l,_) -> 
	Aiger.add_input c l (Aiger.lit2symbol aiger l)
      ) circuit aiger.Aiger.latches
  in
  let rec aux (bdd,circuit) inp = 
    let val_inp = Cudd.bddExistAbstract (Cudd.bddAnd (AigerBdd.bdd_of_variable (AigerBdd.variable_of_input inp)) bdd) cube in
    let circuit, output = AigerBdd.bdd_to_aiger circuit val_inp in
    let output_symbol = Aiger.lit2symbol aiger inp in
    let circuit = Aiger.add_output circuit output output_symbol in
    Cudd.bddAnd bdd (Cudd.bddOr 
		       (Cudd.bddAnd val_inp (AigerBdd.bdd_of_variable (AigerBdd.variable_of_input inp)))
		       (Cudd.bddAnd (Cudd.bddNot val_inp) (Cudd.bddNot (AigerBdd.bdd_of_variable (AigerBdd.variable_of_input inp))))),
    circuit
  in

  let bdd,circuit = List.fold_left aux (strategy,circuit) controllables in
  circuit
  



let test aiger = 
  let contr,uncontr = controllables aiger in
  if !Common.display_debug then List.iter (fun x -> Printf.printf "%d controllable\n" (Aiger.lit2int x)) contr;

  AigerBdd.init aiger;
  let unsafe = List.fold_left Cudd.bddOr (Cudd.bddFalse()) (List.map AigerBdd.bdd_of_variable (List.map AigerBdd.variable_of_output aiger.Aiger.outputs)) in
  let initial = AigerBdd.initial_state aiger in
  let precomputation = AigerBdd.of_aiger aiger in   
  let trap_set = trap precomputation contr uncontr ~weak:false unsafe in
  let trap_set_weak = trap precomputation contr uncontr ~weak:true unsafe in
  let initial_losing = Cudd.bddRestrict (latch_configuration trap_set) initial in
  let initial_losing_weak = Cudd.bddRestrict (latch_configuration trap_set_weak) initial in

  if !Common.display_debug then Cudd.dumpDot ("debug.out") (latch_input_configuration trap_set);

  if Cudd.value initial_losing = 1 
  then print_endline "unrealizable"
  else if Cudd.value initial_losing = 0
  then 
    (print_endline "realizable";
     let strat = strategy_to_aiger aiger (strategy precomputation (negation trap_set)) contr uncontr in
     let file_name = Sys.argv.(1)^".controller.aig" in
     let outch = open_out file_name in
     Aiger.write strat outch;
     close_out outch;
     Printf.printf "controller written in file %s\n" file_name
    )
  else (Cudd.dumpDot ("problem_report.out") initial_losing;  print_endline "problem...");
  
  if Cudd.value initial_losing_weak = 1 
  then print_endline "unrealizable by a weak controller"
  else if Cudd.value initial_losing_weak = 0
  then print_endline "realizable by a weak controller"
  else (Cudd.dumpDot ("problem_report.out") initial_losing_weak;  print_endline "problem...")

let main = 
  if Filename.check_suffix Sys.argv.(0) "attractor" 
     || Filename.check_suffix Sys.argv.(0) "attractor.byte" 
  then 
    if Array.length Sys.argv < 2 
    then Printf.printf "usage : %s <file>\n" Sys.argv.(0)
    else 
      let aiger = Aiger.read_from_file Sys.argv.(1) in
      test aiger
