(* We convert a co-Büchi game into a safety game, where the 
 * goal is to see the target no more than [n] times.
 * [size] is the size of the counter so [n = 2^(size-1)].
 *
 * We take the first output of the given aiger file to be the target state to be avoided.
 * We add an output [safety] that is true when the target has been reached [2^(size-1)] times.
 *)

(* This function adds a counter to count the number of time
 * the target was seen, and an output SAFETY which is true 
 * when the target has been reached [2^(size-1)] times. *)
let to_safety aiger size target =
  let buf = Buffer.create 100 in
  Printf.bprintf buf "output SAFETY;\n";
  Printf.bprintf buf "counter[size=%d](target=%s);\n" size (Aiger.Symbol.to_string target);
  Printf.bprintf buf "SAFETY = counter<%d>;\n" (size-1);
  let dec = ParametricModule.parse_declarations SpeculogModules.env (Buffer.contents buf) in
  ParametricModule.of_declaration (ParametricModule.UseModule(aiger,[]) :: dec)

let to_safety_non_det aiger size target = 
  let buf = Buffer.create 100 in
  Printf.bprintf buf "reg [%d:0] counter;" (size-1);
  Printf.bprintf buf "input [%d:0] NEXT_counter;" (size-1);
  Printf.bprintf buf "reg SAFETY;";
  Printf.bprintf buf "input NEXT_SAFETY;";
  Printf.bprintf buf "input NEXT_OLD_VALID;";
  Printf.bprintf buf "output counter_safety_VALID;";
  (* Printf.bprintf buf "input NEXT_V;";*)
  (* Printf.bprintf buf "input V;"; *)
  (* Printf.bprintf buf "reg VALID;"; *)
  (*Printf.bprintf buf "counter[size=%d](target=%s);\n" size (Aiger.Symbol.to_string target);
  Printf.bprintf buf "xor[size=%d](a=counter,b=NEXT_counter,out=counter_xor);\n" size;
  Printf.bprintf buf "equal[size=%d,value=0](a=counter_xor,out=counter_VALID);\n" size;*)

  Printf.bprintf buf "resize[size_in=1,size_out=%d](a=%s,out=resized_target);\n" size (Aiger.Symbol.to_string target);
  Printf.bprintf buf "adder[size=%d](a=counter,b=resized_target,out=next_value_counter);\n" size;
  Printf.bprintf buf "xor[size=%d](a=next_value_counter,b=NEXT_counter,out=counter_xor);\n" size;
  Printf.bprintf buf "equal[size=%d,value=0](a=counter_xor,out=counter_VALID);\n" size;
  Printf.bprintf buf "counter <- NEXT_counter;\n";

  Printf.bprintf buf "safety_VALID = ! (NEXT_SAFETY ^ counter<%d>);\n" (size-1);
  Printf.bprintf buf "SAFETY <- NEXT_SAFETY;\n";
  Printf.bprintf buf "counter_safety_VALID = (counter_VALID & safety_VALID);\n";
  (*Printf.bprintf buf "VALID <- ( (!INITIALIZED) & counter_VALID & safety_VALID) | (INITIALIZED & V & NEXT_V & counter_VALID & safety_VALID);\n";*)
  let dec = ParametricModule.parse_declarations SpeculogModules.env (Buffer.contents buf) in
  (*let renamed = Aiger.rename aiger [("VALID",None),("V",None)] in*)
  let partial = ParametricModule.of_declaration (ParametricModule.UseModule(aiger,[]) :: dec) in
  let old_valid = match Aiger.lit2tag partial (Aiger.symbol2lit partial ("VALID",None)) with 
    | Latch (l,nl) -> nl 
  in
  let counter_safety_valid = Aiger.symbol2lit partial ("counter_safety_VALID",None) in
  let aiger,new_gate = Aiger.new_var partial in
  let aiger,new_latch = Aiger.new_var aiger in
  let aiger = Aiger.add_and aiger (Aiger.var2lit new_gate) old_valid counter_safety_valid in
  let aiger = Aiger.remove_latch aiger ("VALID",None) in
  (* rename aiger [("VALID",None),("OLD_VALID",None)] in*)

  let aiger = Aiger.add_latch aiger (Aiger.var2lit new_latch) (Aiger.var2lit new_gate) ("VALID",None) in
  Aiger.hide aiger ("counter_safety_VALID",None)



let test file_name size = 
  let aig = Aiger.read_from_file file_name in
  let target = Aiger.lit2symbol aig (List.hd aig.Aiger.outputs) in
  let size = 
    if size = 0
    then 
      (let l = aig.Aiger.num_latches in 
       let s = l + 4 + 1 lsr (l+1) in 
       Printf.eprintf "no size specified using %d\n" s;
       s)
    else size
  in 
  let aig = 
    if List.mem "VALID" (Aiger.names aig) 
    then to_safety_non_det aig size target 
    else to_safety aig size target 
  in
  (* Attractor.test aig*)
  Aiger.write aig stdout

let main = 
  if Filename.basename Sys.argv.(0) = "coBuchi" 
  then   
    if Array.length Sys.argv < 2
    then Printf.printf "usage : %s aiger_file [?size]\n" Sys.argv.(0)
    else
      let size = 
	if Array.length Sys.argv < 3 
	then 0 else int_of_string Sys.argv.(2)
      in
      test Sys.argv.(1) size
		  


  
