(* requiers ocaml graph: install libocamlgraph-ocaml-dev *)
open Graph

module Var = 
  struct 
    type t = Aiger.lit
    let compare = compare 
    let hash = Aiger.lit2int 
    let equal a b = hash a = hash b
  end

module DGraph = Graph.Persistent.Digraph.ConcreteBidirectional(Var)
module Compo = Graph.Components.Make(DGraph)

let aiger_graph m = 
  let tab = Hashtbl.create m.Aiger.maxvar in
  List.iter (fun (s,l,r) -> Hashtbl.add tab s [Aiger.strip l; Aiger.strip r]) m.ands;
  List.iter (fun (l,ll) -> Hashtbl.add tab l [Aiger.strip ll]) m.latches;
  List.iter (fun l -> Hashtbl.add tab l [l]) m.inputs;
  Hashtbl.add tab Aiger.aiger_false [];
  tab


let depends m l =
  let graph = aiger_graph m in
  let insert a = 
    let rec aux accu = function 
      | [] -> List.rev_append accu [a]
      | b :: x when b < a -> aux (b::accu) x
      | b :: x when b > a -> List.rev_append accu (a :: b :: x)
      | b :: x when b = a -> List.rev_append accu (b :: x)
    in aux [] 
  in
  let rec loop accu to_visit = 
    if to_visit = [] then accu
    else 
      let new_accu, new_tv = 
	List.fold_left 
	  (fun (na,ntv) l -> 
	   try 
	     match Hashtbl.find graph l with
	     | [] -> na,ntv
	     | [x] -> (insert l na,ntv)
	     | [l;r] -> (na,(insert l (insert r ntv)))
	   with Not_found -> Printf.printf "\n%d\n" (Aiger.lit2int l); raise Not_found 
	  ) (accu,[]) to_visit
      in loop new_accu new_tv
  in loop [] [Aiger.strip l]

let dependency_graph m = 
  let g = DGraph.empty in
  List.fold_left 
    (fun gr (l,ll) -> 
     let d = depends m ll in 
     List.fold_left (fun gr x -> DGraph.add_edge gr x l) gr d
    ) g (List.map (fun o -> (o,o)) m.outputs @ m.latches)

let reachables g v = 
  let insert l a = 
    let rec aux accu = function 
      | [] -> List.rev_append accu [a]
      | b :: s when b > a -> List.rev_append accu (a :: b :: s)
      | b :: s when b = a -> List.rev_append accu (b :: s)
      | b :: s when b < a -> aux (b::accu) s
    in aux [] l
  in

  let rec loop to_visit visited =
    let ntv = List.fold_left (fun ntv t -> List.fold_left insert ntv (DGraph.succ g t)) [] to_visit in
    let nvis = List.fold_left insert visited to_visit in
    if List.length nvis <> List.length visited
    then loop ntv nvis
    else visited
  in loop [v] []


let names = ref (fun x -> "names undefined")

module M = 
  struct
    type t = DGraph.t
    module V = DGraph.V
    module E = DGraph.E
    let iter_vertex = DGraph.iter_vertex
    let iter_edges_e = DGraph.iter_edges_e
    let graph_attributes i = []
    let default_vertex_attributes i =  []
    let vertex_name v = (!names) v
    let vertex_attributes v = []
    let get_subgraph v = None
    let default_edge_attributes i = []
    let edge_attributes e = [ ]
  end
module GraphPrint = Graph.Graphviz.Dot(M)

let dependency_dot m outch = GraphPrint.output_graph outch (dependency_graph m)

  (*
  Printf.fprintf outch "digraph {\n";
  let name l = match Aiger.get_symbol m l with 
    | Some x -> x
    | None -> string_of_int (Aiger.lit2int l)
  in

  List.iter 
    (fun (l,ll) -> 
     let d = depends m ll in 
     List.iter (fun x -> Printf.fprintf outch "\t%S -> %S;\n" (name l) (name x)) d;
    ) (List.map (fun o -> (o,o)) m.outputs @ m.latches);
  Printf.fprintf outch "}\n\n"
   *)


let usefull m = 
  let visited = Hashtbl.create m.Aiger.maxvar in
  let rec loop to_visit = 
    if to_visit = [] then ()
    else
      let new_to_visit = 
	List.fold_left
	  (fun ntv l ->
	   if Hashtbl.mem visited l then ntv
	   else 
	     match Aiger.lit2tag m l with 
	     | Constant _ -> ntv
	     | Input i -> Hashtbl.replace visited i true; ntv
	     | Latch (l,ll) -> Hashtbl.replace visited l true; (Aiger.strip ll):: ntv
	     | And (s,l,r) ->  Hashtbl.replace visited s true; Aiger.strip l :: Aiger.strip r :: ntv
	  ) [] to_visit
      in loop new_to_visit
  in 
  loop m.outputs;
  (*List.iter (fun o -> loop [o]) m.outputs;*)
  visited


let simplify m = 
  let used = usefull m in
  (* remove unused part of the circuit *)
  let s = Aiger.empty in
  let s = List.fold_left (fun s i -> if Hashtbl.mem used i then Aiger.add_input s i (Aiger.lit2symbol m i) else s) s m.inputs in
  let s = List.fold_left (fun s i -> if Hashtbl.mem used i then Aiger.add_output s i (Aiger.lit2symbol m i) else s) s m.outputs in
  let s = List.fold_left (fun s (l,ll) -> if Hashtbl.mem used l then Aiger.add_latch s l ll (Aiger.lit2symbol m l) else s) s m.latches in
  let s = List.fold_left (fun s (x,l,r) -> if Hashtbl.mem used x then Aiger.add_and s x l r else s) s m.ands in
  s


let decompose m =
  let m1 = Aiger.empty in

  (* Inputs *)
  let m1 = 
    List.fold_left 
      (fun a l -> Aiger.add_input a l (Aiger.lit2symbol m l))
      m1 m.inputs 
  in
  
  (* Latches *)
  let m1 = 
    List.fold_left 
      (fun a (l,ll) -> Aiger.add_latch a l ll (Aiger.lit2symbol m l))
      m1 m.latches
  in

  (* Latches as Outputs *)  
  let m1 = 
    List.fold_left 
      (fun a (l,_) -> Aiger.add_output a l (Aiger.lit2symbol m l))
      m1 m.latches
  in
  
  (* Ands *)
  let m1 = List.fold_left (fun a (s,l,r) -> Aiger.add_and a s l r) m1 m.ands in



  let m2 = Aiger.empty in 

  (* Inputs *)
  let m2 = 
    List.fold_left 
      (fun a l -> Aiger.add_input a l (Aiger.lit2symbol m l))
      m2 m.inputs 
  in
  
  (* Latches as inputs *)
  let m2 = 
    List.fold_left 
      (fun a (l,ll) -> Aiger.add_input a l (Aiger.lit2symbol m l))
      m2 m.latches
  in

  (* Outputs *)  
  let m2 = 
    List.fold_left 
      (fun a l -> Aiger.add_output a l (Aiger.lit2symbol m l))
      m2 m.outputs
  in
  
  (* Ands *)
  let m2 = List.fold_left (fun a (s,l,r) -> Aiger.add_and a s l r) m2 m.ands in

  m1,m2



(* partition the variable to have one input in each part *)
let partition aiger inp =
  let tab = Hashtbl.create 0 in
  let g = dependency_graph aiger in
  let aux a = 
    let r = reachables g a in
    List.iter (fun x -> if not (Hashtbl.mem tab x) then Hashtbl.add tab x a) r
  in 
  List.iter aux inp;
  tab

let partition_list_from_inputs aiger inp = 
  let tab = partition aiger inp in
  let aux a = 
    Hashtbl.fold (fun k i l -> if i = a then k :: l else l) tab []
  in 
  List.rev (List.fold_left (fun p a -> aux a :: p) [] inp)

let partition_list aiger = partition_list_from_inputs aiger aiger.inputs


  

let partition_to_graph aiger tab =
  let dep = dependency_graph aiger in
  DGraph.fold_edges 
    (fun a b g -> 
     try 
       DGraph.add_edge g (Hashtbl.find tab a) (Hashtbl.find tab b)
     with Not_found -> Printf.printf "Not found %d %d\n" (Aiger.lit2int a) (Aiger.lit2int b); g
    ) dep DGraph.empty


(* Restrict the circuit to a given set of inputs and latches *)
let restrict aiger set = 
  let visited = Hashtbl.create aiger.Aiger.maxvar in
  let rec loop to_visit = 
    if to_visit = [] then ()
    else
      let new_to_visit = 
	List.fold_left
	  (fun ntv l ->
	   if Hashtbl.mem visited l then ntv
	   else 
	     match Aiger.lit2tag aiger l with 
	     | Constant _ -> ntv
	     | Input i -> 
		Hashtbl.replace visited i (Aiger.Input i); ntv
	     | Latch (l,ll) -> 
		Hashtbl.replace visited l (Aiger.Latch (l,ll));
		if not (List.mem (Aiger.strip l) set) 
		then ntv 
		else (Aiger.strip ll):: ntv
	     | And (s,l,r) ->  Hashtbl.replace visited s (Aiger.And (s,l,r)); Aiger.strip l :: Aiger.strip r :: ntv
	  ) [] to_visit
      in loop new_to_visit
  in 
  loop set;
  
  (* Warning : we should keep the same identifier for latches and inputs, so that the same BDDs can be reused *)

  let new_uncontr = Hashtbl.create aiger.num_inputs in
  let a = 
    Hashtbl.fold  
      (fun k x a -> 
       match x with 
       | Aiger.Constant _ -> a
       | Input i -> 
	  if not (List.mem (Aiger.strip i) set) 
	  then (Hashtbl.replace new_uncontr k true; a)
	  else Aiger.add_input a k (Aiger.lit2symbol aiger k)

       | Latch (l,ll) -> 
	  if not (List.mem (Aiger.strip l) set) 
	  then (Hashtbl.replace new_uncontr l true; a)
	  else Aiger.add_latch a l ll (Aiger.lit2symbol aiger l)

       | And (s,l,r) -> Aiger.add_and a s l r

      ) visited Aiger.empty 
  in 

  let a =
    Hashtbl.fold 
      (fun k _ a ->
       Aiger.add_input a k (let s = Aiger.lit2symbol aiger k in Printf.sprintf "uncontrollable_%s" (fst s), snd s)
      ) new_uncontr a  

  in

  List.fold_left (fun a b -> Aiger.add_output a b (Aiger.lit2symbol aiger b)) a aiger.outputs







(* Function to test the functionnalities of the module *)
let test () = 
  if Array.length Sys.argv < 2
  then Printf.printf "usage : %s aiger_file\n" Sys.argv.(0);

  let aiger = Aiger.read_from_file Sys.argv.(1) in
(*
  print_endline "=== Dependencies ===";
  let g = dependency_graph aiger in
  names := Aiger.lit2string aiger;
  let f = open_out ("dependency_"^Filename.basename Sys.argv.(1)^".dot") in
  dependency_dot aiger f;
  close_out f;
   
  let sccs = Compo.scc_list g in
  List.iter (fun x ->
	     List.iter (fun y -> Printf.printf " %d" (Aiger.lit2int y)) x;
	     print_newline ()) sccs;


  let display_reach a = 
    let r = reachables g a in
    Printf.printf "=== Reachables from %d ===" (Aiger.lit2int a);
    List.iter (fun y -> Printf.printf " %d" (Aiger.lit2int y)) r;
    print_newline ()
  in
     
  List.iter display_reach aiger.inputs;
   *)

  print_endline "=== Partition ===";
  let c,u = Attractor.controllables aiger in
  let p = partition_list_from_inputs aiger c in 
  (*Hashtbl.iter (fun k i -> Printf.printf "%d -> %d\n" (Aiger.lit2int k) (Aiger.lit2int i)) p;*)
  List.iter (fun x -> List.iter (fun y -> Printf.printf " %d" (Aiger.lit2int y)) x; print_newline ()) p;


  (*
  let f = open_out "partition.dot" in  
  GraphPrint.output_graph f (partition_to_graph aiger (partition aiger));
  close_out f;
  Printf.printf "Wrote partition.dot\n";
 *)

  print_endline "=== Restrict ===";
  Cudd.init (aiger.num_inputs+2*aiger.num_latches+2);

  let initial = Attractor.initial_state aiger in


  let rec loop strat unsafe restriction = function 
    | [] -> strat, unsafe
    (*| [a] -> strat, unsafe*)
    | a :: s ->

       let time = Sys.time () in
       let set = List.rev_append a restriction in 
       Printf.printf "Restricting to : ";
       List.iter (fun y -> Printf.printf " %d" (Aiger.lit2int y)) set; 
       print_newline ();
       let r1 = restrict aiger set in

(*
       print_endline "====== AIGER =>>>>";
       Aiger.write r1 stdout;
       print_endline "<<<<<= AIGER =====";
 *)

       let c,u = Attractor.controllables r1 in
       (*print_endline "controllables"; List.iter (fun x -> Printf.printf "%d controllable\n" (Aiger.lit2int x)) c;*)
    
       print_endline "precompute";
       let p = Attractor.precompute r1 in
       print_endline "value_with_restriction";
       (* we should start with unsafe the states strongly losing of the preceding step *)
       let v = Admissibility.value_with_restriction (strat,Cudd.bdd_true ()) p c u unsafe in
       (*let v = Admissibility.value p c u unsafe in*)
       print_endline "strategies"; 
       let new_strat = Cudd.bddAnd (Admissibility.strategies p v) strat in
       

       (*
       print_endline "checking correctness";
       let correct = Admissibility.correct_strategy p new_strat c u in
       if Cudd.value correct = 1 
       then print_endline "correct strategy"
       else if Cudd.value correct = 0
       then print_endline "uncorrect strategy"
       else print_endline "problem checking correctness...";
	*)


       Printf.printf "size of winning states: %d, cooperatively winning states: %d, losing states: %d\n" (Cudd.dagSize (fst (v Admissibility.Winning))) (Cudd.dagSize (fst (v Admissibility.CoopWinning))) (Cudd.dagSize (fst (v Admissibility.Losing)));

       let initial_value_1 = Cudd.bddRestrict (fst (v Admissibility.Winning)) initial in

       (if Cudd.value initial_value_1 = 1 
	then print_endline "value = 1"
	else 
	   let initial_value_0 = Cudd.bddRestrict (fst (v Admissibility.CoopWinning)) initial in
	   if Cudd.value initial_value_0 = 1 
	   then print_endline "value = 0"
	   else print_endline "value = -1");

       Printf.printf "computation time: %f\n" (Sys.time () -. time);
	 
       loop new_strat (fst (v Admissibility.Losing)) set s

  in 

  let unsafe = Cudd.make_cube (List.map Aiger.lit2int aiger.Aiger.outputs) in


  let all_except_one l =
    let rec aux accu i = function
      | a :: s when i = 0 -> List.rev_append accu s
      | a :: s -> aux (a :: accu) (i-1) s
    in 
    let rec loop accu i = if i < 0 then accu else loop (aux [] i l :: accu) (i-1) in
    loop [] (List.length l - 1)

  in


  let strat, unsafe = 
    List.fold_left (fun (s,u) x -> print_endline "new loop"; loop s u [] x) 
		   (Cudd.bdd_true(),unsafe) 
		   (List.map (fun x -> List.hd p :: x) (all_except_one (List.tl p)))
  in

  (*let strat,unsafe = loop (Cudd.bdd_true ()) unsafe [] p in *)

  
  print_endline "write the strategy obtained after admissibility analysis in file strat.dot";
  Cudd.output_file ("strat.dot") strat;


  let time = Sys.time () in

  (*let unsafe = Cudd.make_cube (List.map Aiger.lit2int aiger.Aiger.outputs) in*)
  (*List.iter (fun x -> Printf.printf "%d controllable\n" (Aiger.lit2int x)) c;
  print_endline "inputs";
  List.iteri (fun i x -> Printf.printf "input %d = %d\n" i (Aiger.lit2int x)) aiger.inputs;*)
  print_endline "precompute";
  let p = Attractor.precompute aiger in

(*  print_endline "admissibility....";
  let v = Admissibility.value_with_restriction (strat,Cudd.bdd_true()) p c u unsafe in
  let strat = Admissibility.strategies_with_restriction (strat,Cudd.bdd_true()) p v in
  print_endline "--->admissibility";
   *)

  let trap_set = Admissibility.check_strategies p strat c u unsafe in
  let initial_losing = Cudd.bddRestrict trap_set initial in

  Printf.printf "check strategy computation time: %f\n" (Sys.time () -. time);

  if Cudd.value initial_losing = 1 
  then 
    (
      print_endline "unsafe strategy";
      print_endline "trying to correct it...";
      let time = Sys.time () in
      let trap_set = Attractor.trap_with_restriction (strat,Cudd.bdd_true()) p c u unsafe in
     
      print_endline "attractor computed";
      let initial_losing = Cudd.bddRestrict trap_set initial in
      Printf.printf "attractor computation time: %f\n" (Sys.time () -. time);

      if Cudd.value initial_losing = 1 
      then print_endline "unrealizable"
      else if Cudd.value initial_losing = 0
      then print_endline "realizable";)
  else if Cudd.value initial_losing = 0
  then print_endline "safe strategy"
  else (Cudd.output_file ("problem_report.dot") initial_losing;  print_endline "problem...")


  

let main = 
  if Filename.basename Sys.argv.(0) = "decomposition" 
  then test ()

