(* Takes an aiger file and determinize it.
 * The idea is to compute the transition relation in the 
 * form of BDDs for each latch, then for each fixed 
 * valuation create a new latch whose transition is given 
 * by restricting the transition relation to this valuation
 * and AIGER_VALID to true.
 *)
let next_string = "NEXT_"
let valid_string = "VALID"
let initialized_string = "INITIALIZED"
(*let () = Common.display_debug := true*)

let valuations_of_lits lit_list = 
    let rec aux = function 
      | [] -> []
      | [x] -> 
	[ Aiger.LitMap.add x true Aiger.LitMap.empty ;
	  Aiger.LitMap.add x false Aiger.LitMap.empty ]

      | hd :: tl -> 
	 let res_tl = aux tl in
	 List.fold_left
	   (fun res partial_valuation ->
	     (Aiger.LitMap.add hd true partial_valuation)
	    :: (Aiger.LitMap.add hd false partial_valuation)
	    :: res
	   ) [] res_tl
    in aux lit_list
    
let reachable_valuations aiger aigerBdd lit_array next_lit_array =
  let valid_lit = Aiger.symbol2lit aiger (valid_string,None) in
  let transition_relation = AigerBdd.composition_vector aigerBdd in
  let valid_bdd = transition_relation.(Aiger.lit2int valid_lit) in
  let initial = AigerBdd.initial_state aiger in
  let cube = AigerBdd.make_cube (Array.to_list lit_array) in

  let successors bdd = 
    let relation = Cudd.bddAnd bdd valid_bdd in
    let next = Cudd.bddExistAbstract relation cube in
    AigerBdd.rename_configuration next next_lit_array lit_array
  in

  let reachables = AigerBdd.fixpoint (fun b -> Cudd.bddOr b (successors b)) initial in
  
  AigerBdd.bdd_to_valuations reachables (Array.to_list lit_array)


let lit2name aiger lit = 
  let symb = Aiger.lit2symbol aiger lit in
  match symb with x,Some i -> x^"_"^string_of_int i | x,None -> x
  
let valuation_to_symbol aiger names v = 
  let buf = Buffer.create 100 in
  (*Aiger.LitMap.iter 
    (fun lit bool ->
    Printf.bprintf buf "%s_%b_" (lit2name aiger lit) bool
    ) v;*)
  List.iter 
    (fun (name,i) ->
      Printf.bprintf buf "%s_%d_" name i
    ) (AigerBdd.extract_valuation aiger v names);
  (Buffer.contents buf, None)

(* replace variables by there next equivalent in valuation *)
let next_valuation aiger valuation = 
  Aiger.LitMap.fold 
    (fun lit bool accu ->
      let (symb_string,symb_option) = Aiger.lit2symbol aiger lit in
      let next_lit = Aiger.symbol2lit aiger (next_string^symb_string,symb_option) in
      Aiger.LitMap.add next_lit bool accu
    ) valuation Aiger.LitMap.empty


(* Here is a schematic description of the procedure:

   i  NEXT_l  NEXT_i             i
   |     |      |                |
  _|_____|______|___           __|________
 |       ___        |         |   _    __ |
 |      |_l_|       |  --->   |  |l|  |!l||    
 |__________________|         |___________|
      |         |                 
    VALID   INITIALIZED
*)
let determinize aiger =

  (* inputs that do not start with next *)
  let inputs = 
    List.filter (fun v -> 
      if String.length v >= (String.length next_string)
      then String.sub v 0 (String.length next_string) <> next_string
      else true
    ) (Aiger.inputs aiger)
  in

  
  let latches = List.filter (fun x -> x <> valid_string && x <> initialized_string) (Aiger.latches aiger) in
  let next_latches = List.map (fun l -> next_string^l) latches in
  (*(Aiger.latches aiger) in*)

  let names_to_lit names = 
    List.fold_left 
      (fun accu name ->
       let literals = Aiger.name_to_literals aiger name in
       Array.fold_left (fun a l -> l :: a) accu literals
      ) [] names
  in

  let precomp = AigerBdd.of_aiger aiger in

  (* List of the possible valuations *)
  (*let valuations = valuations_of_lits (names_to_lit latches) in *)
  let valuations = reachable_valuations aiger precomp (Array.of_list (names_to_lit latches)) (Array.of_list (names_to_lit next_latches)) in

  let lit_initialized = Aiger.symbol2lit aiger (initialized_string,None) in
  let lit_valid = Aiger.symbol2lit aiger (valid_string,None) in
  let valid_bdd = Hashtbl.find (AigerBdd.updates precomp) lit_valid in
  let initialized_bdd = AigerBdd.bdd_of_lit lit_initialized in
  
  Common.debug "adding inputs";
  let a = List.fold_left 
    (fun a l ->
      Aiger.add_input a l (Aiger.lit2symbol aiger l)
    ) Aiger.empty (names_to_lit inputs)
  in

  Common.debug "adding old latches";
  Common.debug "we could remove them at the end";
  let a =
    List.fold_left 
      (fun a (l,nl) -> 
	let symbol = Aiger.lit2symbol aiger l in
	if fst symbol = initialized_string 
	then Aiger.add_latch a l Aiger.aiger_true symbol
	else Aiger.add_latch a l Aiger.aiger_false symbol
      ) a aiger.latches
  in

  (*let a = Aiger.add_latch a lit_initialized Aiger.aiger_true (initialized_string, None) in*)

  Common.debug "adding new latches";
  let tab_latches = Hashtbl.create 100 in
  let a = List.fold_left 
    (fun a v -> 
      let a,vl = Aiger.new_var a in
      Hashtbl.add tab_latches v (Aiger.var2lit vl);
      a
    ) a valuations
  in

  let update_of_valuation valuation = 
    let update = Cudd.bddAnd (AigerBdd.valuation_to_bdd (next_valuation aiger valuation)) valid_bdd in
    update
  in 

  
  (* Converts a state of the subset construction to a set of states of 
     the original game. *)
  let subset_to_states =
    List.fold_left
      (fun accu valuation -> 
	Cudd.bddOr accu 
	  (Cudd.bddAnd 
	     (AigerBdd.valuation_to_bdd valuation)
	     (AigerBdd.bdd_of_lit (Hashtbl.find tab_latches valuation))
	  )
      ) (Cudd.bddFalse()) valuations
  in
    
  let latches_cube = AigerBdd.make_cube (names_to_lit latches) in
  let next_cube = AigerBdd.make_cube (names_to_lit (List.map (fun x -> next_string^x) latches)) in

  let a =
    List.fold_left 
      (fun a v -> 
	let update = Cudd.bddExistAbstract (Cudd.bddAnd initialized_bdd (Cudd.bddAnd (update_of_valuation v) subset_to_states)) latches_cube in
	let update_and_new_latches = Cudd.bddAnd (AigerBdd.valuation_to_bdd (next_valuation aiger v)) update in
	let update_abstracted = Cudd.bddExistAbstract update_and_new_latches next_cube in

	let initialization =
	  Cudd.bddExistAbstract 
	    (Cudd.bddAnd
	       (Cudd.bddNot initialized_bdd)
	       (update_of_valuation v))
	    next_cube
	in
	let update_v = Cudd.bddOr initialization update_abstracted in
	let a,out = AigerBdd.bdd_to_aiger a update_v in
	let a = Aiger.add_latch a (Hashtbl.find tab_latches v) out (valuation_to_symbol aiger latches v) in
	if !Common.display_debug; 
	then Printf.printf "adding latch %s -> %d\n" (Aiger.Symbol.to_string (valuation_to_symbol aiger latches v)) (Aiger.lit2int out);
	  
	a
      ) a valuations
  in

  (* adding outputs *)
  let a =
    List.fold_left 
      (fun a o -> 
	let a,var = Aiger.new_var a in
	let update = Cudd.bddExistAbstract (Cudd.bddAnd initialized_bdd (Cudd.bddAnd (Hashtbl.find (AigerBdd.updates precomp) o) subset_to_states)) latches_cube in
	let a, out = AigerBdd.bdd_to_aiger a update in
	Aiger.add_latch a (Aiger.var2lit var) out (Aiger.lit2symbol aiger o)
	
      ) a aiger.outputs
  in

  a
  

let test file_name = 
  let aiger = Aiger.read_from_file file_name in 
  Cudd.init aiger.Aiger.maxvar;
  Common.debug "determinization test";
  let det = determinize aiger in
  Aiger.write det stdout
    


let main = 
  if Filename.basename Sys.argv.(0) = "determinization" 
  || Filename.basename Sys.argv.(0) = "determinization.byte" 
  then   
    if Array.length Sys.argv < 2
    then Printf.printf "usage : %s aiger_file\n" Sys.argv.(0)
    else test Sys.argv.(1)


