(* The payoff of each player is given by a variable of the circuit *)

(*let () =Common.display_debug := true*)
type player = WeakMullerEquilibria.player
module PlayerMap = WeakMullerEquilibria.PlayerMap

type payoff = string PlayerMap.t

type positive_formula = 
(*| LimInfLE of player * int
| LimSupGE of player * int*)
| LimInfLE of string * int
| LimSupGE of string * int
| Or of positive_formula * positive_formula
| PNeg of negative_formula
and negative_formula =
(*| LimInfGE of player * int
| LimSupLE of player * int*)
| LimInfGE of string * int
| LimSupLE of string * int
| And of negative_formula * negative_formula
| NNeg of positive_formula

let rec positive_formula_to_string = function
  | LimInfLE (s,i) -> Printf.sprintf "liminf %s <= %d" s i
  | LimSupGE (s,i) -> Printf.sprintf "limsup %s >= %d" s i
  | Or (pf1,pf2) -> positive_formula_to_string pf1 ^" | "^ positive_formula_to_string pf2
  | PNeg nf -> "(! "^ negative_formula_to_string nf^")"
and negative_formula_to_string = function
  | LimInfGE (s,i) -> Printf.sprintf "liminf %s >= %d" s i
  | LimSupLE (s,i) -> Printf.sprintf "limsup %s <= %d" s i
  | And (nf1,nf2) -> negative_formula_to_string nf1 ^" & "^ negative_formula_to_string nf2
  | NNeg nf -> "(! "^ positive_formula_to_string nf^")"

type formula = (Coalition.t * positive_formula list * negative_formula list) list

let formula_to_string fl = 
  let aux accu (c,pf,nf) = 
    (
      List.fold_left (fun string f -> string ^" & "^positive_formula_to_string f) 
		     (List.fold_left (fun string f -> string ^" & "^negative_formula_to_string f) ("["^Coalition.to_string c^"] ") nf)
		     pf
    ) ^ accu
  in List.fold_left aux "" fl

type positive_negative_formula = 
| PF of positive_formula
| NF of negative_formula
    
exception ParseError of string 

let parse_positive_negative_formula stream = 
  let rec parse = parser   
    | [< 'Genlex.Kwd "liminf"; 'Genlex.Ident p; res = parse_after_liminf p >] -> res
    | [< 'Genlex.Kwd "limsup"; 'Genlex.Ident p; res = parse_after_limsup p >] -> res
    | [< 'Genlex.Kwd "!"; f = parse; res = parse_rest
	(match f with PF g -> (NF (NNeg g)) | NF g -> (PF (PNeg g))) >] -> 
      res

    | [< 'Genlex.Kwd "("; f = parse; 'Genlex.Kwd ")" >] -> f
      | [< >] -> raise (ParseError "waiting liminf <int> or limsup <int> or !")
  and parse_after_liminf p = parser
      | [< 'Genlex.Kwd "<="; 'Genlex.Int i; res = parse_rest_positive (LimInfLE (p,i)) >] -> res
      | [< 'Genlex.Kwd ">="; 'Genlex.Int i; res = parse_rest_negative (LimInfGE (p,i)) >] -> res
      | [< >] -> raise (ParseError "waiting <= or >=")
  and parse_after_limsup p = parser
      | [< 'Genlex.Kwd ">="; 'Genlex.Int i; res = parse_rest_positive (LimSupGE (p,i)) >] -> res
      | [< 'Genlex.Kwd "<="; 'Genlex.Int i; res = parse_rest_negative (LimSupLE (p,i)) >] -> res
      | [< >] -> raise (ParseError "waiting <= or >=")
  and parse_rest_positive accu = parser
      | [< 'Genlex.Kwd "|"; f = parse >] -> 
	(match f with PF g -> PF (Or (accu,g)) | _ -> failwith "cannot have | in positive formula")
      | [< >] -> PF accu
  and parse_rest_negative accu = parser
      | [< 'Genlex.Kwd "&"; f = parse >] -> 
	(match f with NF g -> NF (And (accu,g)) | _ -> failwith "cannot have & in negative formula")
      | [< 'Genlex.Kwd "|" >] -> failwith "cannot have | in positive formula"
      | [< >] -> NF accu
  and parse_rest accu = match accu with 
    | PF g -> parse_rest_positive g
    | NF g -> parse_rest_negative g
  in parse stream

let parse_formula stream = 

  let rec parse_positive_negative_list = 
    parser
| [< f = parse_positive_negative_formula; x = parse_remaining_list 
    (match f with PF g -> [g],[] | NF g -> [],[g]) >] -> x
  and parse_remaining_list (pf,nf) = 
    parser
      | [< 'Genlex.Kwd "&"; x = parse_positive_negative_formula; y = parse_remaining_list 
	  (match x with PF g -> (g::pf,nf) | NF g -> (pf,g::nf)) >] -> y
      | [< >] -> (pf,nf)
  in

  let rec parse_players = parser 
    | [< 'Genlex.Int p ; c = parse_players_remaining (Coalition.singleton p)>] -> c
    | [< >] -> Coalition.empty

  and parse_players_remaining accu = parser
      | [< 'Genlex.Kwd ";"; 'Genlex.Int p ; s = parse_players_remaining (Coalition.add p accu) >] -> s
      | [< >] -> accu
  in

  let rec parse = parser 
| [< 'Genlex.Kwd "["; p = parse_players; 'Genlex.Kwd "]"; pf,nf = parse_positive_negative_list; n2 = parse_remainder [p,pf,nf] >] -> n2
    | [< 'Genlex.Kwd "("; p = parse; 'Genlex.Kwd ")"; n2 = parse_remainder p >] -> n2
    | [< >] -> failwith "parse formula failed: waiting \"[players] payoff formula\""
  and parse_remainder accu = parser
      | [< 'Genlex.Kwd "&"; n2 = parse >] -> List.rev_append n2 accu
      | [< >] -> accu
       
  in 
  try parse stream 
  with x -> 
    
    Printf.eprintf "unexpected token:%s\n" (Common.next_token stream);
    raise x


let lexer = Genlex.make_lexer ["liminf";"limsup";"<=";">="; "&"; "|"; "!";"[";"]";";";"(";")"]

let formula_of_string string = 
  let stream = Stream.of_string string in
  parse_formula (lexer stream)

let rec pf2af = function
  | LimInfLE (p,i) -> 
      StateFormula.of_smaller_equal p (*PlayerMap.find p payoff*) i
  | LimSupGE (p,i) -> 
    StateFormula.of_greater_equal p (*PlayerMap.find p payoff*) i
  | Or (pf1,pf2) -> 
    let a = pf2af pf1 in
    let b = pf2af pf2 in
    StateFormula.disjunction a b
  | PNeg nf -> nf2af nf 
    
and nf2af = function
  | LimInfGE (p,i) -> 
    StateFormula.of_greater_equal p (*PlayerMap.find p payoff*) i
  | LimSupLE (p,i) -> 
    StateFormula.of_smaller_equal p (*PlayerMap.find p payoff*) i
  | And (nf1,nf2) -> 
    let a = nf2af nf1 in
    let b = nf2af nf2 in
    StateFormula.conjunction a b
  | NNeg pf -> pf2af pf

    
let positive_formula_to_atomic_formula, negative_formula_to_atomic_formula = pf2af,nf2af 


let not_deviator x = StateFormula.AF_equal ((Translation.deviator x),0)
let is_deviator x = StateFormula.AF_equal ((Translation.deviator x),1)

let deviators_included players coalition = 
  let elements = Coalition.elements (Coalition.diff players coalition) in
  List.fold_left StateFormula.conjunction (StateFormula.of_bool true) (List.map not_deviator elements)

let deviators_not_included players coalition = 
  let elements = Coalition.elements (Coalition.diff players coalition) in
  List.fold_left StateFormula.disjunction (StateFormula.of_bool false) (List.map is_deviator elements)

let formula_to_Buchi_coBuchi aiger (*payoff*) players formula =
  (*let players = PlayerMap.fold (fun p i c -> Coalition.add p c) payoff Coalition.empty in*)

  let coalition_formula (accu_buchi,accu_coBuchi) (c,pf_list,nf_list) =
    let buchi_list = List.fold_left 
      (fun accu pf ->
	let atomic = positive_formula_to_atomic_formula (*payoff*) pf in
	let deviators = deviators_included players c in
	let form = StateFormula.implication deviators atomic in
	print_string "buchi : ";
	print_endline (StateFormula.to_string form);
	form :: accu
      ) accu_buchi pf_list
    in
    let coBuchi = List.fold_left
      (fun accu nf ->
	let atomic = negative_formula_to_atomic_formula nf in
	let deviators = deviators_included players c in
	let form = StateFormula.conjunction (StateFormula.negation atomic) deviators in
	StateFormula.disjunction form accu
      ) accu_coBuchi nf_list
    in
    (buchi_list,coBuchi)
  in

  let buchi_list, coBuchi = 
    List.fold_left coalition_formula ([], StateFormula.of_bool false) formula
  in

  Printf.printf "coBuchi: %s\n"(StateFormula.to_string coBuchi);
  Common.debug "coBuchi:";
  Common.debug (StateFormula.to_string coBuchi);


  let trans = Translation.of_aiger aiger in
  let deviator_game,contr,uncontr = Translation.deviator_game trans in

  AigerBdd.init deviator_game;
  let coBuchiBdd = StateFormula.to_bdd (Aiger.name_to_literals deviator_game) coBuchi in
  (*Cudd.dumpDot "coBuchi.dot" coBuchiBdd;
  print_endline "dumped coBuchi.dot";*)

  (deviator_game,
   contr,
   uncontr,
   List.map (StateFormula.to_bdd (Aiger.name_to_literals deviator_game)) buchi_list,
   coBuchiBdd)


let formula_to_parity aiger players formula = 
  let aiger, contr, uncontr, buchi_list,coBuchi = formula_to_Buchi_coBuchi aiger players formula in
  (* We have to add an observer for the Buchi conditions *)
  (*let aiger_inputs = 
    List.fold_left (fun a x -> Aiger.add_input a x (Aiger.lit2symbol aiger x)) Aiger.empty aiger.Aiger.inputs 
  in*)
  
  let aiger,nb,condition_list = 
    List.fold_left 
      (fun (aiger,i,list) bdd -> 
	let a,lit = AigerBdd.bdd_to_aiger aiger bdd in
	
	let condition_i = Printf.sprintf "condition_%d" i in
	let a = Aiger.add_output a lit (condition_i,None) in
	if !Common.display_debug then (Aiger.write a stdout);
	(*Environment.add_module env condition_i *)
	a, i+1, condition_i :: list
      ) (aiger,0,[]) buchi_list
  in

  let aiger,parity2 =
    if nb = 0
    then 
      let aiger = Aiger.add_output aiger Aiger.aiger_true ("PARITY2",None) in 
      aiger, Cudd.bddTrue ()
    else
      let spec =
	(* we have to be carefull about the case nb= 1 as well *)
	if nb = 1 
	then
	  let buf = Buffer.create 100 in
	  Printf.bprintf buf "output PARITY2;\n";
	  Printf.bprintf buf "PARITY2 = condition_0;\n";
	  Buffer.contents buf
	else
	  let size = Common.log nb in
	  
	  let buf = Buffer.create 100 in
	  if size = 1 
	  then 
	    Printf.bprintf buf "reg BUCHI_OBSERVER;\n"
	  else
	    Printf.bprintf buf "reg [%d:0] BUCHI_OBSERVER;\n" (size-1);
	  Printf.bprintf buf "output PARITY2;\n";
	  for i = 0 to nb - 1 do
	    Printf.bprintf buf "assign[size=%d,value=%d](out=v%d);\n" size i i;
	    Printf.bprintf buf "equal[size=%d,value=%d](a=BUCHI_OBSERVER,out=bo_equals_%d);\n" size i i;
	  done;
	  for i = 0 to nb - 1 do
	    Printf.bprintf buf "c%d = (bo_equals_%d & ! condition_%d) | (bo_equals_%d & condition_%d);\n" i i i (if i = 0 then nb-1 else i-1 ) (if i = 0 then nb-1 else i-1 )
	  done;
	  
	  Printf.bprintf buf "switch[nb=%d,size=%d](" nb size;
	  for i = 0 to nb - 1 do
	    Printf.bprintf buf "case%d=c%d," i i;
	    Printf.bprintf buf "then%d=v%d," i i;
	  done;
	  Printf.bprintf buf "out=new_bo);\n";
	  Printf.bprintf buf "BUCHI_OBSERVER <- new_bo;\n";
	  Printf.bprintf buf "PARITY2 = (bo_equals_%d & condition_%d);\n" (nb-1) 0;
	  Buffer.contents buf

      in
      print_endline spec;

      let declarations = SpeculogParser.parse_declarations SpeculogModules.env spec in
	
      let aiger =
	ParametricModule.of_declaration (UseModule (aiger,[]) :: declarations)
      in 
      
      let aiger =
	List.fold_left (fun aiger c -> Aiger.hide aiger (c,None)) aiger condition_list
      in aiger , 
	 AigerBdd.bdd_of_gate (Aiger.symbol2lit aiger ("PARITY2",None))
  (*StateFormula.conjunction (StateFormula.of_equal "bo_equals"nb-1) (StateFormula.of_equal "condition_0" 1)*)
  in

  let aiger,parity3lit = AigerBdd.bdd_to_aiger aiger coBuchi in
  let aiger = Aiger.add_output aiger parity3lit ("PARITY3",None) in

  aiger, 
  contr,
  uncontr,
  parity2,
  coBuchi
    

let equilibria_region aiger players formula =
  let aiger,contr,uncontr,priority2,priority3 = formula_to_parity aiger players formula in
  let parity_file = open_out "parity.aig" in
  Aiger.write aiger parity_file;
  close_out parity_file;
  print_endline "wrote parity.aig";
  (* Common.display_debug := true;*)

  let contr,uncontr = Attractor.controllables aiger in  (* useless ? *)
(*  let priority2 = AigerBdd.bdd_of_variable (Aiger.symbol2lit aiger ("parity2",None)) in
  let priority3 = AigerBdd.bdd_of_variable (Aiger.symbol2lit aiger ("parity3",None)) in *)
  let game = AigerBdd.of_aiger aiger in

  let win_weak = Parity.parity3 game contr uncontr priority2 priority3 true (Attractor.tt()) in
  aiger,win_weak 

let test aiger_file formula_string = 
  let aiger = Aiger.read_from_file aiger_file in
  let formula = formula_of_string formula_string in
  let players = Coalition.of_aiger aiger in
  print_endline "there is a problem with the values return by formula_to_parity";
  let aiger,win = equilibria_region aiger players formula in
  let initial = AigerBdd.initial_state aiger in

  let initial_winning = Cudd.bddRestrict (Attractor.latch_configuration win) initial in

  if Cudd.value initial_winning = 1 
  then print_endline "realizable"
  else if Cudd.value initial_winning = 0
  then print_endline "unrealizable"
  else (Cudd.dumpDot ("problem_report.out") initial_winning;  print_endline "problem...")

let main = 
  if Filename.basename Sys.argv.(0) = "limitEquilibria" ||
    Filename.basename Sys.argv.(0) = "limitEquilibria.byte"
  then
    if Array.length Sys.argv < 3
    then
      Printf.printf "usage : %s aiger_file formula\n" Sys.argv.(0)
    else
      test Sys.argv.(1) Sys.argv.(2)

