(** Solves a Buchi game *)
let parity2 game controllables uncontrollables priority2 weak region =
  let rec aux region =     
    Common.debug "parity2 aux";
    let a = Attractor.attractor_with_restriction game controllables uncontrollables ~weak (Attractor.of_bdds priority2 (Cudd.bddFalse())) region in
    let b = Attractor.trap_with_restriction game controllables uncontrollables ~weak (Attractor.diff region a) region in
    let new_region = Attractor.diff region b in
    if Attractor.equal new_region region
    then region
    else aux new_region
  in aux region


(** Solve a game with 3 priorities. 
   The second argument is the set of states that have priority 2.
   The third argument is the set of states that have priority 3.
   Other states are assumed to have priority 1. *)
let parity3 game controllables uncontrollables priority2 priority3 weak region =
  let rec aux region = 
    Common.debug "parity3 aux";
    let trap = Attractor.trap_with_restriction game controllables uncontrollables ~weak (Attractor.of_bdds priority3 (Cudd.bddFalse())) region in
    let win = parity2 game controllables uncontrollables priority2 weak (Attractor.diff region trap) in
    let attractor = Attractor.attractor_with_restriction game controllables uncontrollables ~weak win region in
    let new_region = Attractor.diff region attractor in
    if Attractor.equal new_region region
    then attractor
    else
      let win2 = aux new_region in
      Attractor.union win2 attractor
  in aux region

let () = Common.display_debug := true


let test aiger =
  let contr,uncontr = Attractor.controllables aiger in

  AigerBdd.init aiger;
  let priority2 = 
    try AigerBdd.bdd_of_gate (Aiger.symbol2lit aiger ("PARITY2",None)) 
    with Not_found -> 
      AigerBdd.bdd_of_gate (Aiger.symbol2lit aiger ("FAIRNESS",None)) 
  in
  let priority3 = 
    try AigerBdd.bdd_of_gate (Aiger.symbol2lit aiger ("PARITY3",None)) 
    with Not_found -> 
      print_endline "warning: no output PARITY3";
      Cudd.bddFalse ()
  in
  let initial = AigerBdd.initial_state aiger in
  Common.debug "computing game";
  let game = AigerBdd.of_aiger aiger in   
  Common.debug "computing winning region";
  let win = 
    if Cudd.equal priority3 (Cudd.bddFalse())
    then 
      parity2 game contr uncontr priority2 false (Attractor.tt()) 
    else
      if Cudd.equal priority3 (Cudd.bddTrue())
      then Attractor.ff()
      else
	parity3 game contr uncontr priority2 priority3 false (Attractor.tt())
  in
  
  let win_weak = 
    if Cudd.equal priority3 (Cudd.bddFalse())
    then 
      parity2 game contr uncontr priority2  true (Attractor.tt())
    else
      if Cudd.equal priority3 (Cudd.bddTrue())
      then Attractor.ff ()
      else
	parity3 game contr uncontr priority2 priority3 true (Attractor.tt())
  in

  Cudd.dumpDot "win.dot" (Attractor.latch_configuration win);
  let initial_winning = Cudd.bddRestrict (Attractor.latch_configuration win) initial in
  let initial_weak_winning = Cudd.bddRestrict (Attractor.latch_configuration win_weak) initial in

  (*Cudd.dumpDot ("winning.dot") win;  
  Cudd.dumpDot ("initial.dot") initial;  *)

  if Cudd.value initial_winning = 1 
  then print_endline "realizable"
  else if Cudd.value initial_winning = 0
  then print_endline "unrealizable"
  else (Cudd.dumpDot ("problem_report.dot") initial_winning;  print_endline "problem...");
  if Cudd.value initial_weak_winning = 1 
  then print_endline "realizable by a weak controller"
  else if Cudd.value initial_weak_winning = 0
  then print_endline "unrealizable by a weak controller"
  else (Cudd.dumpDot ("problem_report.dot") initial_weak_winning;  print_endline "problem...")

let main = 
  if Filename.check_suffix Sys.argv.(0) "parity" 
    || Filename.check_suffix Sys.argv.(0) "parity.byte" 
  then 
    if Array.length Sys.argv < 2 
    then Printf.printf "usage : %s <file>\n" Sys.argv.(0)
    else 
      let aiger = Aiger.read_from_file Sys.argv.(1) in
      test aiger
