module StringSet = Set.Make(String)

module Proposition = 
struct 
  type t = Var of string * int | And of t * t | Or of t * t | Not of t | True | False

  let rec eval valuation = function
    | Var (s,i) -> valuation s = i
    | And (a,b) -> eval valuation a && eval valuation b
    | Or (a,b) -> eval valuation a || eval valuation b
    | Not a -> not (eval valuation a)
    | True -> true
    | False -> false

  let rec parse = 
    parser
    | [< a = parse_atomic; r = parse_prop_remainder a >] -> r
    | [< 'Genlex.Kwd "("; e = parse; 'Genlex.Kwd ")"; r = parse_prop_remainder e >] -> r
    | [< 'Genlex.Kwd "!"; e = parse_prop_not; r = parse_prop_remainder e >] -> r
  and parse_atomic = 
      parser
    | [< 'Genlex.Ident name >]
      (*'Genlex.Kwd "=" ; 'Genlex.Int i >] *) -> Var(name,1)
    | [< 'Genlex.Kwd "true" >] -> True
    | [< 'Genlex.Kwd "false" >] -> False
  and parse_prop_remainder e =
    parser
    | [< 'Genlex.Kwd "|";  f = parse >] -> Or(e,f)
    | [< 'Genlex.Kwd "&";  f = parse >] -> And(e,f)
    | [< >] -> e
  and parse_prop_not =
    parser
    | [< 'Genlex.Ident name >]
      (* ; 'Genlex.Kwd "=" ; 'Genlex.Int i >]*) -> (Var(name,0))
    | [< 'Genlex.Kwd "("; e = parse; 'Genlex.Kwd ")" >] -> Not e

  let parse_keywords = ["=";"(";")";"!";"|";"&";"true";"false"]

  let rec to_string = function
    | Var (s,i) ->
       if i = 1 then s else if i = 0 then "(!"^s^")" else failwith "comparison with integers impossible"
    | And (a,b) -> "("^to_string a ^" & "^to_string b^")"
    | Or (a,b) -> "("^to_string a ^" | "^to_string b^")"
    | Not a -> "(!("^to_string a ^"))"
    | True -> " 1 "
    | False -> " 0 "

  let compare a b = compare a b

  let rec labels = function
    | Var (s,i) -> StringSet.singleton s
    | And (a,b) -> StringSet.union (labels a) (labels b)
    | Or (a,b) -> StringSet.union (labels a) (labels b)
    | Not a -> labels a
    | True | False -> StringSet.empty

  let to_speculog prop = 
    let rec aux = function 
      | Var (s,i) -> 
	 if i = 1 then s else if i = 0 then "(!"^s^")" else failwith "comparison with integers impossible"
      | And (a,b) -> "("^aux a ^" & "^aux b^")"
      | Or (a,b) -> "("^aux a ^" | "^aux b^")"
      | Not a -> "(!("^aux a ^"))"
      | True -> " 1 "
      | False -> " 0 "
    in aux prop

  let to_bdd aiger prop = 
    let rec aux = function
      | Var (s,i) -> 
	 (*Printf.printf "to_bdd: using lit %d for symbol %s\n" (Aiger.lit2int (Aiger.symbol2lit aiger (s,None))) s;*)
	 if i = 1
	 then AigerBdd.bdd_of_gate (Aiger.symbol2lit aiger (s,None)) 
	 else if i = 0
	 then AigerBdd.bdd_of_not_gate (Aiger.symbol2lit aiger (s,None)) 
	 else failwith "comparison with integers impossible"
      | And (a,b) -> Cudd.bddAnd (aux a) (aux b)
      | Or (a,b) ->  Cudd.bddOr (aux a) (aux b)
      | Not a -> Cudd.bddNot (aux a) 
      | True -> Cudd.bddTrue()
      | False -> Cudd.bddFalse()
    in aux prop

  let conj a b = match a,b with 
    | True , y | y , True -> y
    | False , _ | _ , False -> False
    | x , y -> And (x,y)

  let disj a b = match a,b with 
    | True , _ | _ , True -> True
    | False , y | y , False -> y
    | x , y -> Or (x,y)

  let neg a = match a with 
    | True -> False | False -> True | Not x -> x | x -> Not x
end

type t = Prop of Proposition.t | Alt of t * t | Concat of t * t | Star of t | Epsilon | Empty

let rec accepts_epsilon = function
  | Alt (a,b) -> accepts_epsilon a || accepts_epsilon b
  | Concat (a,b) -> accepts_epsilon a && accepts_epsilon b
  | Star a -> true
  | Epsilon -> true
  | _ -> false

let alt a b = 
  if compare a b = 0 then a
  else
    match a,b with 
    | Empty, x -> x
    | x,Empty -> x
    | Epsilon, y | y , Epsilon when accepts_epsilon y -> y
    | x,y -> Alt(x,y)

let concat a b = match a,b with
  | Empty, _ -> Empty
  | _, Empty -> Empty
  | Epsilon,x -> x
  | x,Epsilon -> x
  | x,y -> Concat (x,y)

let star = function
  | Epsilon -> Epsilon
  | Empty -> Epsilon
  | x -> Star x

let rec to_string = function
  | Alt (a,b) -> "("^to_string a ^"|"^to_string b^")"
  | Concat (a,b) ->  "( ("^to_string a ^") ("^to_string b^") )"
  | Star a -> "("^to_string a ^"*)"
  | Epsilon -> " epsilon "
  | Empty -> " nothing "
  | Prop p -> " { "^Proposition.to_string p^" } " 

  
let residual valuation = 
  let rec aux = function
    | Prop p ->      
      if Proposition.eval valuation p 
      then Epsilon
      else Empty
    | Epsilon -> Empty
    | Empty -> Empty
    | Alt (a,b) -> alt (aux a) (aux b)
    | Concat (a,b) -> 
      let ra = concat (aux a) b in
      if accepts_epsilon a 
      then alt ra (aux b)
      else ra
    | Star a -> concat (aux a) (star a)
  in aux 

module PropMap = Map.Make(Proposition)

let residuals = 
  let rec aux = function
    | Prop p ->      
       let map = PropMap.add p Epsilon PropMap.empty in
       PropMap.add (Proposition.neg p) Empty map
    | Epsilon -> PropMap.add Proposition.True Empty PropMap.empty
    | Empty -> PropMap.add Proposition.True Empty PropMap.empty
    | Alt (a,b) -> 
       let map_a = aux a in 
       let map_b = aux b in
       PropMap.fold 
	 (fun p e map -> 
	  PropMap.fold 
	    (fun q f map ->
	     PropMap.add (Proposition.conj p q) (alt e f) map
	    ) map_b map
	 ) map_a PropMap.empty
	 
    | Concat (a,b) -> 
       let map_a = aux a in
       let map_b = aux b in 
       PropMap.fold 
	 (fun q f map ->
	  let ra = concat f b in
	  if accepts_epsilon a
	  then PropMap.fold 
		 (fun p e map -> 
		  let conj = Proposition.conj p q in
		  PropMap.add conj (alt ra e) map
		 ) map_b map
	  else PropMap.add q ra map
	 ) map_a PropMap.empty
	 
    | Star a -> 
       let map_a = aux a in
       PropMap.fold 
	 (fun q f map ->
	  PropMap.add q (concat f (star a)) map
	 ) map_a PropMap.empty

  in aux 

let rec parse = 
  parser 
  | [< 'Genlex.Kwd "{"; p = Proposition.parse; 'Genlex.Kwd "}"; e = parse_remainder (Prop p) >] -> e
  | [< 'Genlex.Kwd "("; e = parse; 'Genlex.Kwd ")"; f = parse_remainder e >] -> f
  | [< 'Genlex.Kwd "."; e = parse_remainder (Prop (Proposition.True)) >] -> e
  | [< >] -> Epsilon
and parse_remainder e = 
  parser
  | [< 'Genlex.Kwd "|"; p = parse >] -> alt e p
  | [< 'Genlex.Kwd "*"; p = parse >] -> concat (star e) p
  | [< 'Genlex.Kwd "+"; p = parse >] -> concat (concat e (star e)) p
  | [< 'Genlex.Kwd "?"; p = parse >] -> alt (concat e p) p
  | [< 'Genlex.Int i; p = parse >] -> 
    let rec loop i = if i = 1 then e else concat e (loop (i-1)) in 
    concat (loop i) p
  | [< p = parse >] -> concat e p

let parse_keywords = List.rev_append ["{";"}";".";"?";"+";"*"] Proposition.parse_keywords

let of_string s = parse (Genlex.make_lexer parse_keywords (Stream.of_string s))

type expr = t
module Expr = 
  struct 
    type t = expr
    let compare a b = compare a b 
  end

module Set = Set.Make(Expr)
module Int = struct type t = int let compare = compare end
module IntMap = Map.Make(Int)
module Map = Map.Make(Expr)

let rec labels = function 
  | Prop p -> Proposition.labels p
  | Alt (a,b) | Concat (a,b) -> StringSet.union (labels a) (labels b)
  | Star a -> labels a
  | Epsilon | Empty -> StringSet.empty

let all_residuals expr = 
  let rec loop seen expr =
    if Set.mem expr seen then seen
    else
      (
	let seen = Set.add expr seen in
	let map = residuals expr in
	PropMap.fold (fun p e seen -> loop seen e) map seen
      )
  in loop Set.empty expr

(* the argument [res_map] maps residuals to some integers *)
let transitions res_map = 
  Map.fold
    (fun expr index map -> 
      let residual = residuals expr in
      PropMap.fold 
	(fun p e map -> 
	  (* Printf.eprintf "expr: %s -> prop: %s -> expr:%s\n" (to_string expr) (Proposition.to_string p) (to_string e); *)
	  let index_e = Map.find e res_map in
	  try
	    IntMap.add index_e (Proposition.disj (Proposition.conj p (Proposition.Var ("RES_"^string_of_int index, 1))) (IntMap.find index_e map)) map
	  with Not_found -> IntMap.add index_e (Proposition.conj p (Proposition.Var ("RES_"^string_of_int index, 1))) map
	) residual map
    ) res_map IntMap.empty


let to_speculog expr =
  let res = all_residuals expr in
  let res_map,nb = Set.fold (fun r (map,i) -> Map.add r i map, i+1) (Set.remove expr res) (Map.singleton expr 0,1) in
  let transitions = transitions res_map in

  let size = Common.log (nb+1) in
  let labels = labels expr in

  let buf = Buffer.create 500 in
  StringSet.iter (fun s -> Printf.bprintf buf "input %s;\n" s) labels;
  Printf.bprintf buf "reg [%d:0] RESIDUAL;\n" (size-1);
  (* Printf.bprintf buf "output ACCEPT;\n";*)
  (*Printf.bprintf buf "assign[size=%d,value=%d](out=residual0);\n" size (nb+1);*)
  Map.iter (fun expr index -> 
    (* Printf.eprintf "expr: %s -> %d\n" (to_string expr) index; *)
    Printf.bprintf buf "equal[size=%d,value=%d](a=RESIDUAL,out=RESIDUAL_equal_%d);\n" size index index) res_map;
  IntMap.iter 
    (fun int prop -> 
      (* Printf.eprintf "prop: %d -> %s\n" int (Proposition.to_string prop); *)
     Printf.bprintf buf "assign[size=%d,value=%d](out=v%d);\n" size int int;
     Printf.bprintf buf "cond%d = %s;\n" int (Proposition.to_speculog prop);
     Printf.bprintf buf "if[size=%d](condition=cond%d,then=v%d,else=residual%d,out=residual%d);\n" size int int int (int+1)
    ) transitions;
  Printf.bprintf buf "RESIDUAL <- residual%d;\n" nb;
  
  let acceptation =
    Map.fold 
      (fun expr index accu -> 
       if accepts_epsilon expr 
       then Proposition.disj accu (Proposition.Var ("RESIDUAL_equal_"^string_of_int index, 1))
       else accu
      ) res_map Proposition.False
  in
  Printf.bprintf buf "ACCEPT = %s;\n" (Proposition.to_speculog acceptation);
  print_endline "speculog generated:\n";
  print_endline (Buffer.contents buf);
  Buffer.contents buf

  

let to_aiger expr =
  (*Common.display_debug := true;*)
  let aiger = Aiger.empty in
  let labels = labels expr in
  let aiger =
    StringSet.fold
      (fun i aiger -> 
       let aiger,v = Aiger.new_var aiger in
       (*Printf.printf "adding input %d\n" (Aiger.lit2int (Aiger.var2lit v));*)
       Aiger.add_input aiger (Aiger.var2lit v) (i,None)
      ) labels aiger
  in 
  let res = all_residuals expr in
  let res_map,nb = Set.fold (fun r (map,i) -> Map.add r i map, i+1) (Set.remove expr res) (Map.singleton expr 0,1) in
  let size = Common.log (nb-1) (*nb+1*) in

  let aiger = 
    Map.fold (fun r index aiger ->
      let aiger,v = Aiger.new_var aiger in
      let aiger = Aiger.add_input aiger (Aiger.var2lit v) ("RES_"^string_of_int index, None) in
      (*Printf.printf "adding input %d\n" (Aiger.lit2int (Aiger.var2lit v));*)
      aiger
    ) res_map aiger 
  in

  let transitions = transitions res_map in
  let env = 
    IntMap.fold
      (fun int prop env -> 
       let bdd = Proposition.to_bdd aiger prop in
       (*let aiger,v = Aiger.new_var aiger in*)
       let aiger,cond = AigerBdd.bdd_to_aiger aiger bdd in
       let aiger = Aiger.add_output aiger cond ("out",None) in
       (*Printf.printf "\nCONDITION %d\n" int;
       Aiger.write aiger stdout; *)
       Environment.add_module env ("CONDITION_"^string_of_int int) aiger
      
    ) transitions SpeculogModules.env
  in

  
  let buf = Buffer.create 500 in
  StringSet.iter (fun s -> Printf.bprintf buf "input %s;\n" s) labels;
  Printf.bprintf buf "reg [%d:0] RESIDUAL;\n" (size-1);
  Printf.bprintf buf "output ACCEPT;\n";
  Printf.bprintf buf "assign[size=%d,value=%d](out=residual0);\n" size (nb+1); 

  Map.iter (fun expr index -> 
    Printf.bprintf buf "equal[size=%d,value=%d](a=RESIDUAL,out=RESIDUAL_equal_%d);\n" size index index) res_map;
  
  let renaming () = 
    Map.iter (fun expr index -> Printf.bprintf buf ",RES_%d=RESIDUAL_equal_%d" index index) res_map
  in
    
(*  Printf.bprintf buf "CONDITION_0(out=cond0";
  renaming ();
  Printf.bprintf buf ");\n";*)

  let rec loop index =
    try 
      let prop = IntMap.find index transitions in
      Printf.bprintf buf "assign[size=%d,value=%d](out=v%d);\n" size index index;
      Printf.bprintf buf "CONDITION_%d(out=cond%d" index index;
      renaming ();
      Printf.bprintf buf ");\n";
      Printf.bprintf buf "if[size=%d](condition=cond%d,then=v%d,else=residual%d,out=residual%d);\n" size index index index (index+1);
      if index < nb - 1 then loop (index+1) else ()
    with Not_found -> 
      Printf.bprintf buf "resize[size_in=%d,size_out=%d](a=residual%d,out=residual%d);\n" size size index (index+1);
      if index < nb - 1 then loop (index+1) else ()
  in loop 0;

  Printf.bprintf buf "RESIDUAL <- residual%d;\n" nb;
  let acceptation =
    Map.fold 
      (fun expr index accu -> 
       if accepts_epsilon expr 
       then Proposition.disj accu (Proposition.Var ("RESIDUAL_equal_"^string_of_int index, 1))
       else accu
      ) res_map Proposition.False
  in
  Printf.bprintf buf "ACCEPT = %s;\n" (Proposition.to_speculog acceptation);

  
  Common.debug (Buffer.contents buf);
  SpeculogParser.parse_module env (Buffer.contents buf)
  

let module_regexp string =
  let expr = of_string string in 
  Cudd.init 100;
  Common.debug (to_string expr);
  (*Common.display_debug := true;*)
  let aiger = to_aiger expr in
  Cudd.quit ();
  aiger

let env = 
  Environment.add_parametric_module SpeculogModules.env "regexp" (fun f -> module_regexp (ParametricModule.get_string (f "expr"))) 

let () = 
  Environment.current_env := (fun () -> env)


let test string = 
  Aiger.write (module_regexp string) stdout

let main = 
  if Filename.basename Sys.argv.(0) = "regularExpression" 
  || Filename.basename Sys.argv.(0) = "regularExpression.byte" 
  then
    if Array.length Sys.argv < 2
    then
      Printf.printf "usage : %s regexp\n" Sys.argv.(0)
    else
      test Sys.argv.(1);


