let product_renaming a i = 
  let i_string = string_of_int i in
  (* we rename the outputs and the latches in order to distinguish the names *)
  let renaming = List.map (fun x -> x, x^"__"^i_string) (List.rev_append (Aiger.outputs a) (Aiger.latches a)) in
  (* we want the controllable inputs to be duplicated (with the second version uncontrollable) 
   * but the uncontrollable inputs stay the same *)
  let renaming2 = List.map
		    (fun x -> 
		     x, if Attractor.controllable_name x 
			then if i = 1 then x^"__"^i_string 
			     else "uncontrollable_"^x^i_string
			else x
		    ) (Aiger.inputs a) 
  in List.rev_append renaming renaming2



let display_renaming r = 
  List.iter 
    (fun (x,lit) -> 
     Printf.printf "%s  " x;
     Array.iteri (fun i x -> Printf.printf " %d:%d" i (Aiger.lit2int x)) lit;
     print_newline ()
    ) r


let simulation_game a b =
  let renaming_a = product_renaming a 1 in
  let renaming_b = product_renaming b 2 in
  let renamed_a = Aiger.full_rename a renaming_a in
  let renamed_b = Aiger.full_rename b renaming_b in
  let c = Aiger.compose renamed_a renamed_b in

  let renaming_lit x i = 
    let i_string = string_of_int i in
    List.map (fun x -> x, Aiger.name_to_literals c (x^"__"^i_string)) (List.rev_append (Aiger.outputs x) (Aiger.latches x))
  in

  c, renaming_lit a 1, renaming_lit b 2



(* the simulation relation is the set of pairs (s,t) from which there is a winning strategy in the simulation game *)
let simulation aig = 
  let c,r1,r2 = simulation_game aig aig in
  let contr, uncontr = Attractor.controllables c in

  (* unsafe states are the one that are losing for the first component but not the second *)
  let unsafe = match Aiger.outputs c with 
    | a :: b :: [] -> 
       if a.[String.length a - 1] = '1' && b.[String.length b - 1] = '2' 
       then 
	 Cudd.bddAnd (Cudd.bddNot (Cudd.ithVar (Aiger.lit2int (Aiger.symbol2lit c (b,None))))) (Cudd.ithVar (Aiger.lit2int (Aiger.symbol2lit c (a,None))))
       else if a.[String.length a - 1] = '2' && b.[String.length b - 1] = '1' 
       then Cudd.bddAnd (Cudd.bddNot (Cudd.ithVar (Aiger.lit2int (Aiger.symbol2lit c (a,None))))) (Cudd.ithVar (Aiger.lit2int (Aiger.symbol2lit c (b,None))))
       else failwith "there is not one output for 1 and one output for 2 as expected..."
    | _ -> failwith "there are not two outputs, I don't know what the objective should be"
  in
  let precomputation = Attractor.precompute c in   
  Attractor.trap precomputation contr uncontr unsafe, r1 , r2



let test () = 
  if Array.length Sys.argv < 2 
  then Printf.printf "usage : %s <file>\n" Sys.argv.(0);
  let aig = Aiger.read_from_file Sys.argv.(1) in
  Cudd.init (aig.Aiger.num_inputs+2*aig.Aiger.num_latches+2);
  let c,r1,r2 = simulation aig in
  display_renaming r1;
  display_renaming r2;
  Cudd.dumpDot "simulation.dot" c;
  print_endline "simulation relation written in file simulation.dot"

let main =
  if Filename.check_suffix Sys.argv.(0) "simulationRelation" 
  then test ()  
