(* parsing of (a subset of) the smv language *)
type expr = 
| True
| False
| Var of string            (* boolean variable, e.g., s = TRUE *)
| NextVar of string
| Int of string * int      (* e.g. s = 12 *) 
| NextInt of string * int
| Neg of expr
| And of expr * expr
| Or of expr * expr
| Imp of expr * expr
| Eq of expr * expr

(** Type of a SMV module.*)
type t = {
  vars : string list;                   (* boolean variables *)
  ints : (string * int list) list;
  init : expr list;
  trans : expr list;
  fair : expr list;
}

let empty = { vars=[]; ints=[]; init=[]; trans=[]; fair=[] }

let parse_expr x = 
  let rec simple = parser
	   | [< 'Genlex.Kwd "TRUE" >] -> True
	   | [< 'Genlex.Kwd "FALSE" >] -> False
	   | [< 'Genlex.Kwd "next"; 'Genlex.Kwd "("; 'Genlex.Ident s; 
		'Genlex.Kwd ")"; e = after_next s >] -> e
	     (* we have to be carefull about the priorities *)
	   | [< 'Genlex.Kwd "!"; e = simple >] -> Neg e
	   | [< 'Genlex.Ident s; e = after_ident s >] -> e
	   | [< 'Genlex.Kwd "("; s = aux; 'Genlex.Kwd ")" >] -> s
	   | [< >] -> 
	     failwith "unexpected token: waiting 'TRUE' 'FALSE' 'next' '(' '!(' or an identifier"
	       (*
  and after_not = parser 
      | [< 'Genlex.Kwd "("; e = aux; 'Genlex.Kwd ")" >]  -> e
      | [< 'Genlex.Ident s >]  -> Var s*)
  and after_ident s = parser
      | [< 'Genlex.Kwd "="; 'Genlex.Int i >] ->  Int (s, i) 
      | [<  >] -> Var s
  and after_next s = parser
      | [< 'Genlex.Kwd "="; 'Genlex.Int i; >] ->  NextInt (s, i)
      | [<  >] -> NextVar s

  and parse_cont accu = parser 
      | [< 'Genlex.Kwd "&" ; e2 = aux >] ->  And(accu,e2)
      | [< 'Genlex.Kwd "|" ; e2 = aux >] ->  Or(accu,e2)
      | [< 'Genlex.Kwd "->" ; e2 = aux >] ->  Imp(accu,e2)
      | [< 'Genlex.Kwd "<->" ; e2 = aux >] ->  Eq(accu,e2)
      | [<  >] -> accu
  and aux = parser 
      | [< a = simple; b = parse_cont a >] -> b
      | [< >] -> 
	failwith "unexpected token: waiting TRUE FALSE etc..."
  in 
  aux x


let parse_vars x = 
  let rec parse_int_set_remaining accu = 
    parser
    | [< 'Genlex.Kwd ","; 'Genlex.Int i; l = parse_int_set_remaining (i::accu) >] -> l
    | [< >] -> accu
  in
  let parse_int_set = 
    parser 
    | [< 'Genlex.Int i ; l = parse_int_set_remaining [i] >] -> l
    | [< >] -> failwith "unexpected token: waiting integer"
  in

  let parse_type (vars,ints) v = 
    parser 
    | [< 'Genlex.Kwd "boolean" >] -> (v::vars,ints)
    | [< 'Genlex.Kwd "{"; l = parse_int_set; 'Genlex.Kwd "}" >] -> (vars,(v,l) :: ints)
  in
  let rec vars accu = parser 
	 | [< 'Genlex.Ident v; 'Genlex.Kwd ":"; t = parse_type accu v; 'Genlex.Kwd ";" ; v = vars t >] -> v
	 | [< >] ->  accu
  in 
  vars ([],[]) x 


let parse_module =
  let rec aux name m = parser
	 | [< 'Genlex.Kwd "MODULE"; 'Genlex.Ident n; 
	      res = aux (Some n) m >] -> res

	 | [< 'Genlex.Kwd "VAR"; vars,ints = parse_vars; 
	      res = aux name {m with vars=List.rev_append vars m.vars; ints=List.rev_append ints m.ints} >] -> 
	   res

	 | [< 'Genlex.Kwd "INIT"; init = parse_expr; 
	      res = aux name {m with init=init :: m.init} >] -> res
	     
	 | [< 'Genlex.Kwd "TRANS"; trans = parse_expr; 
	      res = aux name {m with trans=trans :: m.trans} >] -> res

	 | [< 'Genlex.Kwd "FAIRNESS"; fair = parse_expr; 
	      res = aux name {m with fair=fair :: m.fair} >] -> res
	 
	 | [< >] -> name, m
  in aux None empty

let read inch = 
  let lexer = Genlex.make_lexer ["TRUE";"FALSE";"next";"="; "("; ")";"|";"!";"&";"->";"<->";"boolean";"VAR";"INIT";"TRANS";"FAIRNESS";"MODULE";":";"{";"}";",";";"]  in
  
  let char_stream = Stream.of_channel inch in
  let ignore_comment = 
    let rec aux i = 
      match Stream.peek char_stream with
      | Some p when p = '-' ->
	(
	  Stream.junk char_stream;
	  match Stream.peek char_stream with
	  | Some q when q = '-' ->
	    while Stream.peek char_stream <> Some '\n'
	    do Stream.junk char_stream done;
	    Some (Stream.next char_stream)
	  | _ -> Some '-'
	)
      | x -> Stream.junk char_stream; x
    in Stream.from aux 
  in

  let stream = lexer ignore_comment in

  try parse_module stream
  with Stream.Error x -> 
    Printf.printf "unexpected token %s\n" (Common.next_token stream);
    raise (Stream.Error x)


let conversion int_list int = 
  let rec aux i = function
    | [] -> failwith "conversion not found"
    | hd :: tl when hd = int -> i
    | hd :: tl -> aux (i+1) tl
  in aux 0 int_list

  
let conversion_by_name name ints int = 
  try conversion (List.assoc name ints) int
  with Not_found -> 
    Printf.eprintf "warning: %s was not found among integer variables\n" name;
    raise Not_found

(* We convert the [next(s)] expressions to [NEXT_s].
 * Also convert operators to Speculog syntax.
 * [extras] contains extra variable that are not declared as var in the smv file *)
let rec expr_to_string m equals extras expr = 
  let rec aux accu extras = function
    | True -> accu,extras,"true"
    | False -> accu,extras, "false"
    | Var s -> 
       if List.mem s m.vars || List.exists (fun (a,_) -> a = s) m.ints || List.mem s extras
       then accu, extras, s else accu, s::extras, s
						     
    | NextVar s -> 
       let ns = "NEXT_"^s in
       if List.mem s m.vars || List.exists (fun (a,_) -> a = s) m.ints || List.mem ns extras
       then accu, extras, ns
       else accu, ns::extras, ns

    | Int (s, i) -> 
      let converted = conversion_by_name s m.ints i in
      let size = Common.log (List.length (List.assoc s m.ints)) in
      let equal = (size,converted,s) in
      (if List.mem equal accu then accu else equal :: accu),
      extras,
      "equal_"^string_of_int converted^"_"^s

    | NextInt (s, i) -> 
      let converted = conversion_by_name s m.ints i in
      let size = Common.log (List.length (List.assoc s m.ints)) in
      let equal = (size,converted,"NEXT_"^s) in
      (if List.mem equal accu then accu else equal :: accu),
      extras,
      "equal_"^string_of_int converted^"_NEXT_"^s

    | Neg e -> let a,e,s = aux accu extras e in a,e,"(!"^s^")"
    | And (e1, e2) -> 
      let accu,ex1,se1 = aux accu extras e1 in
      let accu,ex2,se2 = aux accu ex1 e2 in
      accu, ex2, "("^se1^" & "^se2^")"
    | Or (e1, e2) -> 
      let accu,ex1,se1 = aux accu extras e1 in
      let accu,ex2,se2 = aux accu ex1 e2 in
      accu,ex2, "("^se1^" | "^se2^")"
    | Imp (e1, e2) -> 
      let accu,ex1,se1 = aux accu extras e1 in
      let accu,ex2,se2 = aux accu ex1 e2 in
      accu,ex2, "((! "^se1^") | "^se2^")"
    | Eq (e1, e2) -> 
       let accu,ex1,se1 = aux accu extras e1 in
       let accu,ex2,se2 = aux accu ex1 e2 in
      accu, ex2, "! ("^se1^" xor "^se2^")"
  in aux equals extras expr
									   
let conj_expr_to_string m equals extras list = match list with 
  | [] -> equals, extras,"true"
  | hd :: tl -> 
    List.fold_left 
       (fun (accu,extras,string) expr -> 
	 let naccu, nex, s = expr_to_string m accu extras expr in
	 naccu, nex, "("^string ^") & ("^s^")"
       ) (expr_to_string m equals extras hd) tl

let rec expr_next = function
| Var x -> NextVar (x)
| Int (s,i) -> NextInt (s,i)
| NextVar _ -> failwith "next should not appear in initialization"
| NextInt _ -> failwith "next should not appear in initialization"
| Neg e -> Neg (expr_next e)
| And (a,b) -> And(expr_next a,expr_next b)
| Or (a,b) -> Or(expr_next a,expr_next b)
| Imp (a,b) -> Imp(expr_next a,expr_next b)
| Eq (a,b) -> Eq(expr_next a,expr_next b)
| x -> x

let to_speculog outch name smv = 
  (match name with Some x -> Printf.fprintf outch "module %s;\n" x | _ -> ());
  List.iter (fun v -> Printf.fprintf outch "reg %s;\n" v) smv.vars;
  List.iter (fun (v,list) -> Printf.fprintf outch "reg [%d:0] %s;\n" (Common.log (List.length list) - 1) v) smv.ints;
  List.iter (fun v -> Printf.fprintf outch "input NEXT_%s;\n" v) smv.vars;
  List.iter (fun (v,list) -> Printf.fprintf outch "input [%d:0] NEXT_%s;\n" (Common.log (List.length list) - 1) v) smv.ints;

  let intermediary_init, extras1, expr_init = conj_expr_to_string smv [] [] (List.map expr_next smv.init) in
  let intermediary_trans, extras2, expr_trans = conj_expr_to_string smv intermediary_init extras1 smv.trans in
  let intermediary_fair, extras3, expr_fairness = 
    List.fold_left
      (fun (inter,extra,exprs) f -> 
	let inter,extra,expr = expr_to_string smv inter extra f in 
	inter,extra,expr::exprs
      ) (intermediary_trans,extras2,[]) smv.fair
  in

  List.iter (fun s -> Printf.fprintf outch "input %s;\n" s) extras3;
  let nb_fairness = List.fold_left (fun i v -> Printf.fprintf outch "output FAIRNESS_%d;\n" i; i+1) 0 expr_fairness in
  Printf.fprintf outch "reg INITIALIZED;\n";
  Printf.fprintf outch "reg VALID;\n";
  Printf.fprintf outch "INITIALIZED <- 1;\n";

  List.iter 
    (fun (size,value,name) -> 
      Printf.fprintf outch "equal[size=%d,value=%d](a=%s, out=equal_%d_%s);\n" size value name value name
    ) intermediary_fair;

  let nb_fairness = List.fold_left (fun i e -> Printf.fprintf outch "FAIRNESS_%d = %s;\n" i e; i+1) 0 expr_fairness in

  Printf.fprintf outch "VALID <- ((!INITIALIZED) & (%s)) | (INITIALIZED & (%s));\n" 
    expr_init expr_trans;

  List.iter (fun v -> Printf.fprintf outch "%s <- NEXT_%s;\n" v v) smv.vars;
  List.iter (fun (v,list) -> Printf.fprintf outch "%s <- NEXT_%s;\n" v v) smv.ints;
  (match name with Some x -> Printf.fprintf outch "endmodule\n" | _ -> ())
    

let test file_name = 
  let file = open_in file_name in
  let name,smv = read file in
  close_in file;
  to_speculog stdout None smv 

let main = 
  if Filename.check_suffix Sys.argv.(0) "smv" 
  then 
    if Array.length Sys.argv < 2 
    then Printf.printf "usage : %s <file>\n" Sys.argv.(0)
    else 
      test Sys.argv.(1) 



