module StringSet = Set.Make(String)

type t = 
| AF_true 
| AF_false
| AF_or of t * t
| AF_and of t * t
| AF_not of t
| AF_equal of string * int
    
let to_string = 
  let rec aux = function
    | AF_true -> "true"
    | AF_false -> "false"
    | AF_or (a,b)-> Printf.sprintf "(%s or %s)" (aux a) (aux b)
    | AF_and (a,b)-> Printf.sprintf "(%s and %s)" (aux a) (aux b)
    | AF_not (a)-> Printf.sprintf "(not %s)" (aux a)
    | AF_equal (s,i) -> Printf.sprintf "(%s = %d)" s i
  in aux

let to_name = 
  let rec aux = function
    | AF_true -> "true"
    | AF_false -> "false"
    | AF_or (a,b)-> Printf.sprintf "(%s_or_%s)" (aux a) (aux b)
    | AF_and (a,b)-> Printf.sprintf "(%s_and_%s)" (aux a) (aux b)
    | AF_not (a)-> Printf.sprintf "(not_%s)" (aux a)
    | AF_equal (s,i) -> Printf.sprintf "(%s_=_%d)" s i
  in aux

let to_short_string = 
  let rec aux = function
    | AF_true -> "true"
    | AF_false -> "false"
    | AF_or (a,b)-> Printf.sprintf "(%s|%s)" (aux a) (aux b)
    | AF_and (a,b)-> Printf.sprintf "(%s&%s)" (aux a) (aux b)
    | AF_not (a)-> Printf.sprintf "(!%s)" (aux a)
    | AF_equal (s,i) -> Printf.sprintf "(%s=%d)" s i
  in aux
  
let compare (x:t) (y:t) = compare x y

let rec to_bdd name_to_lit = 
  let rec aux = function
    | AF_true -> Cudd.bddTrue()
    | AF_false -> Cudd.bddFalse()
    | AF_or (a,b) -> Cudd.bddOr (aux a) (aux b)
    | AF_and (a,b) -> Cudd.bddAnd (aux a) (aux b)
    | AF_not a -> Cudd.bddNot (aux a)
    | AF_equal (name,int) -> 
      let literals = name_to_lit name in
      if Array.length literals = 0 
      then (Printf.eprintf "warning: no litteral found for variable %s\n" name; Cudd.bddTrue())
      else
      let i,bdd = Array.fold_left
	(fun (i,bdd) lit -> 
	  let eq =
	    if (int land (1 lsl i)) > 0 
	    then Cudd.ithVar (Aiger.lit2int lit)
	    else Cudd.bddNot (Cudd.ithVar (Aiger.lit2int lit))
	  in
	  (i+1,Cudd.bddAnd bdd eq)
	) (0,Cudd.bddTrue ()) literals
      in bdd
  in aux 

let parse stream =
  let rec parse_atomic = parser
| [< 'Genlex.Ident name; 'Genlex.Kwd "=" ; 'Genlex.Int i >] -> (AF_equal(name,i))
| [< 'Genlex.Kwd "true" >] -> AF_true 
| [< 'Genlex.Kwd "false" >] -> AF_false
| [< 'Genlex.Kwd "("; tf=parse_temp; 'Genlex.Kwd ")" >] -> tf
| [< 'Genlex.Kwd "not"; tf = parse_temp >] -> AF_not(tf)
| [< 'Genlex.Kwd "!"; tf = parse_temp >] -> AF_not(tf)
  and parse_temp_remainder r = parser
      | [< 'Genlex.Kwd "or"; s=parse_temp >] -> AF_or(r,s)
      | [< 'Genlex.Kwd "|"; s=parse_temp >] -> AF_or(r,s)
      | [< 'Genlex.Kwd "and"; s=parse_temp >] -> AF_and(r,s)
      | [< 'Genlex.Kwd "&"; s=parse_temp >] -> AF_and(r,s)
      | [< 'Genlex.Kwd "->"; s=parse_temp >] -> AF_or(s,AF_not r)
      | [< 'Genlex.Kwd "implies"; s=parse_temp >] -> AF_or(s,AF_not r)
      | [< >] -> r
  and parse_temp = parser 
      | [< af = parse_atomic ; tf = parse_temp_remainder af >] -> tf
  in parse_atomic stream

let tokens = ["true";"false";"or";"|";"not";"!";"and";"&"; "("; ")";"=";"->";"implies"]  
let lexer = Genlex.make_lexer tokens

let of_string string =
  let stream = lexer (Stream.of_string string) in
  parse stream

let rec names = function
  | AF_or (a,b) | AF_and (a,b) -> StringSet.union (names a) (names b)
  | AF_not a -> names a
  | AF_equal (name,i) -> StringSet.singleton name
  | _ -> StringSet.empty

let of_bool b = if b then AF_true else AF_false
let of_equal x i = AF_equal (x,i)
  
let conjunction a b = match a,b with
  | AF_true, x | x, AF_true -> x
  | AF_false, x | x , AF_false -> AF_false
  | x,y -> AF_and (x,y)
let disjunction a b = match a,b with 
  | AF_true, x | x, AF_true -> AF_true
  | AF_false, x | x , AF_false -> x
  | x, y -> AF_or (x,y)
let negation a = match a with 
  | AF_true -> AF_false
  | AF_false -> AF_true 
  | AF_not a -> a
  | a -> AF_not a
let implication a b = match a,b with 
  | AF_true, x -> x 
  | x, AF_false -> negation x
  | x,y -> disjunction (negation x) y


  (* warning this function may not be optimized *)
let rec of_smaller_equal x i = 
  if i = -1 then AF_false else disjunction (AF_equal (x,i)) (of_smaller_equal x (i-1))

let of_greater_equal x i = negation (of_smaller_equal x (i-1))
