(* Translation from multiplayer games to two player games in the aiger format.
   Name of the inputs should begin with the player that controls it.
   For instance, an input named [player_1_a] is controlled by player 1.
   In case they are no such comment inputs are assumed to belong to player 0.

*)




type player = Coalition.Player.t
module Player = Coalition.Player

type t = {
  aiger:Aiger.t; 
  players:Coalition.t;
  controller:Aiger.lit -> Player.t;
}

let add_player g p = {g with players = Coalition.add p g.players}

let of_aiger aiger = 
  let control = Hashtbl.create aiger.Aiger.num_inputs in
  let controller input = Hashtbl.find control input in
  
  let players =
    List.fold_left
      (fun players input ->
	let lits = Aiger.name_to_literals aiger input in
	let b = match Str.split (Str.regexp "[_]") input with
	  | "player" :: d :: _ -> 
	    (try int_of_string d
	     with e -> 
	       Printf.eprintf "the syntax is : input_<int>_name\n";
	       raise e)
	  | _ ->
	    Printf.eprintf "warning: %s controller by player 0\n" input; 0
	in 	      
	Array.iter (fun l -> Hashtbl.add control l b) lits;
	Coalition.add b players
      ) Coalition.empty (Aiger.inputs aiger)
  in

  {aiger=aiger;players=players;controller=controller}

let read_from_file file =
  let aiger = Aiger.read_from_file file in
  of_aiger aiger

let deviator player = Printf.sprintf "deviator_%d" player


let deviator_game mgame =  
  let game = Aiger.empty in
  let tab_uncontrol = Hashtbl.create mgame.aiger.num_inputs in

  (* Inputs controled by Adam. If the original input was 
     named [i] there is now an input named [uncontrollable_i].  *)
  let game = List.fold_left
    (fun game input -> 
      let game,new_var = Aiger.new_var game in
      let new_lit = Aiger.var2lit new_var in
      let (name,i) = Aiger.lit2symbol mgame.aiger input in
      Hashtbl.add tab_uncontrol input new_lit;
      Aiger.add_input game new_lit ("uncontrollable_"^name,i)
    ) game mgame.aiger.inputs 
  in

  (* Eve can suggest the input. For each input [i] of the 
     originial game, we add an input named [controllable_i]. *)
  (*let tab_control = Hashtbl.create mgame.aiger.num_inputs in *)
  let game,ci = List.fold_left 
    (fun (game,ci) input -> 
      let game,new_var = Aiger.new_var game in
      let new_lit = Aiger.var2lit new_var in
      let (name,i) = Aiger.lit2symbol mgame.aiger input in
      Aiger.add_input game new_lit ("controllable_"^name,i), 
      new_lit::ci
    ) (game,[]) mgame.aiger.inputs 
  in

  (* Gates to compute deviators *)
  let tab_deviator_gate  = Hashtbl.create (Coalition.cardinal mgame.players) in

  (* A gate for each player that is true when inputs for eve and adam differ *)
  (* input control i != input uncontrol i  (ci & !ui || !ci & ui) *)
  let game =       
    List.fold_left 
      (fun game input ->

	let game,new_var_1 = Aiger.new_var game in
	let game,new_var_2 = Aiger.new_var game in
	let game,new_var_3 = Aiger.new_var game in
	let new_lit_1 :: new_lit_2 :: new_lit_3 :: [] = List.map Aiger.var2lit [new_var_1;new_var_2;new_var_3] in

	(*let ci = Hashtbl.find tab_uncontrol input in
	let ui = Hashtbl.find tab_control_latch input in*)
	let (name,i) = Aiger.lit2symbol mgame.aiger input in
	let ui = Aiger.symbol2lit game ("uncontrollable_"^name,i) in
	(*let ci = Aiger.symbol2lit game ("input_"^name,i) in*)
	let ci = Aiger.symbol2lit game ("controllable_"^name,i) in

 	(* first gate computes [g1 = ci & !ui] *)
	let game = Aiger.add_and game new_lit_1 ci (Aiger.aiger_not ui) in
	(* second gate computes [g2 = !ci & ui] *)
	let game = Aiger.add_and game new_lit_2 ui (Aiger.aiger_not ci) in
	(* third gate computes [g3 = !g1 & !g2] true if inputs are the same*)
	let game = Aiger.add_and game new_lit_3 (Aiger.aiger_not new_lit_1) (Aiger.aiger_not new_lit_2) in

	try 
	  let old_gate = Hashtbl.find tab_deviator_gate (mgame.controller input) in
	  (* the fourth literal is the disjunction of the old 
	     gate and g3 [g4 = ! (!old & g3)] *)
	  let game,new_var_4 = Aiger.new_var game in
	  let new_lit_4 = Aiger.var2lit new_var_4 in
	  Hashtbl.add tab_deviator_gate (mgame.controller input) (Aiger.aiger_not new_lit_4);
	  Aiger.add_and game new_lit_4 (Aiger.aiger_not old_gate) new_lit_3
	  
	with Not_found -> 
	  Hashtbl.add tab_deviator_gate (mgame.controller input) (Aiger.aiger_not new_lit_3);
	  game

      ) game mgame.aiger.inputs
  in

  (* Latches to remember deviators *)
  let tab_deviator_latch = Hashtbl.create (Coalition.cardinal mgame.players) in

  let game = Coalition.fold
    (fun player game -> 
      let game, new_var = Aiger.new_var game in
      let new_lit = Aiger.var2lit new_var in
      let game, new_gate_var = Aiger.new_var game in
      let new_gate = Aiger.var2lit new_gate_var in
      let gate = 
	try Hashtbl.find tab_deviator_gate player 
	with Not_found -> Aiger.aiger_false 
      in
      let game = Aiger.add_and game new_gate (Aiger.aiger_not new_lit) (Aiger.aiger_not gate) in
      Aiger.add_latch game new_lit (Aiger.aiger_not new_gate) (Printf.sprintf "deviator_%d" player,None)
    ) mgame.players game 
  in


  (* Normal latch of the initial game *)
  let tab_latches = Hashtbl.create (List.length mgame.aiger.latches) in
  let game = 
    List.fold_left
      (fun game (l,n) -> 
	let game,new_var = Aiger.new_var game in
	Hashtbl.add tab_latches l (Aiger.var2lit new_var);
	game
      ) game mgame.aiger.latches
  in

  (* Gates discribing the transitions *)
  let tab_gates = Hashtbl.create (List.length mgame.aiger.ands) in
  let game = 
    List.fold_left
      (fun game (l,_,_) -> 
	let game,new_var = Aiger.new_var game in
	Hashtbl.add tab_gates l (Aiger.var2lit new_var);
	game
      ) game mgame.aiger.ands
  in

  let find x = 
    try
      let y = match Aiger.lit2tag mgame.aiger (Aiger.strip x) with 
      | Constant b -> (Aiger.strip x)
      | And (lit,_,_) -> Hashtbl.find tab_gates lit
      | Latch (lit,_) -> Hashtbl.find tab_latches lit
      | Input lit -> Hashtbl.find tab_uncontrol lit 
      in if Aiger.sign x then Aiger.aiger_not y else y
    with Not_found -> 
      Printf.printf "literal %d was not found\n" (Aiger.lit2int x);
      Printf.printf "= literal %s\n" (Aiger.Symbol.to_string (Aiger.lit2symbol mgame.aiger x));
      raise Not_found
  in 

  (* adding latches of the original game *)
  let game = List.fold_left 
    (fun game (l,next) ->
      let (name,i) = Aiger.lit2symbol mgame.aiger l in
      Aiger.add_latch game (find l) (find next) (name,i)
    ) game mgame.aiger.latches
  in

  (*print_endline "adding ands of the original game";  *)
  let game = List.fold_left 
    (fun game (l,r0,r1) -> 
      Aiger.add_and game (find l) (find r0) (find r1)
    ) game mgame.aiger.ands
  in

(*
  let error_gate = ref Aiger.aiger_false in
  let game = List.fold_left
    (fun game player ->
      let new_lit = new_var () in
      let game = Aiger.add_and game new_lit (find (mgame.objective player)) (Hashtbl.find tab_deviator_latch player) in
      let new_lit_1 = new_var () in
      let game = Aiger.add_and game new_lit_1 (Aiger.aiger_not new_lit) (Aiger.aiger_not !error_gate) in
      error_gate := Aiger.aiger_not new_lit_1;
      game
    ) game mgame.players
  in
*)
  (* adding outputs *)
  let game = List.fold_left (fun game o -> 
    let sym = Aiger.lit2symbol mgame.aiger o in
    Aiger.add_output game (find o) sym
  ) game mgame.aiger.outputs
  in

  let ui = Hashtbl.fold (fun k i l -> i :: l) tab_uncontrol [] in
  (*let ci = Hashtbl.fold (fun k i l -> i :: l) tab_control [] in*)

  game, ci, ui




let test file = 
  let g = read_from_file file in
  let d,contr,uncontr = deviator_game g in
  Aiger.write d stdout

let main = 
  if Filename.basename Sys.argv.(0) = "translation" ||
    Filename.basename Sys.argv.(0) = "translation.byte"
  then
    if Array.length Sys.argv < 2
    then
      Printf.printf "usage : %s aiger_file\n" Sys.argv.(0)
    else
      test Sys.argv.(1) 


