module StringSet = StateFormula.StringSet

module Formula =
struct
  type temporal_formula = 
  | TF_eventually of StateFormula.t
  | TF_always of StateFormula.t
      
  type t = 
  | F_or of t * t
  | F_and of t * t
  | F_implies of t * t
  | F_not of t
  | F_temp of temporal_formula


  let to_string = 
    let rec aux = function
      | F_not s -> Printf.sprintf "(not %s)" (aux s)    
      | F_or(a,b) -> Printf.sprintf "(%s or %s)" (aux a) (aux b)
      | F_and(a,b) -> Printf.sprintf "(%s and %s)" (aux a) (aux b)
      | F_implies(a,b) -> Printf.sprintf "(%s implies %s)" (aux a) (aux b)
      | F_temp(TF_eventually af) -> Printf.sprintf "(eventually %s)" (StateFormula.to_string af)
      | F_temp(TF_always af) -> Printf.sprintf "(always %s)" (StateFormula.to_string af)
    in aux

  let to_name = 
    let rec aux = function
      | F_not s -> Printf.sprintf "(not_%s)" (aux s)    
      | F_or(a,b) -> Printf.sprintf "(%s_or_%s)" (aux a) (aux b)
      | F_and(a,b) -> Printf.sprintf "(%s_and_%s)" (aux a) (aux b)
      | F_implies(a,b) -> Printf.sprintf "(%s_implies_%s)" (aux a) (aux b)
      | F_temp(TF_eventually af) -> Printf.sprintf "(eventually_%s)" (StateFormula.to_name af)
      | F_temp(TF_always af) -> Printf.sprintf "(always_%s)" (StateFormula.to_name af)
    in aux

  let to_short_string = 
    let rec aux = function
      | F_not s -> Printf.sprintf "(!%s)" (aux s)    
      | F_or(a,b) -> Printf.sprintf "(%s|%s)" (aux a) (aux b)
      | F_and(a,b) -> Printf.sprintf "(%s&%s)" (aux a) (aux b)
      | F_implies(a,b) -> Printf.sprintf "(%s => %s)" (aux a) (aux b)
      | F_temp(TF_eventually af) -> Printf.sprintf "(F %s)" (StateFormula.to_short_string af)
      | F_temp(TF_always af) -> Printf.sprintf "(G %s)" (StateFormula.to_short_string af)
    in aux


  let parse string =
    let lexer = Genlex.make_lexer (["true";"false";"or";"|";"not";"!";"and";"&";"implies";"=>";"eventually";"F";"always";"G";"="; "("; ")"; "end"] @ StateFormula.tokens) in
    let stream = lexer (Stream.of_string string) in
    
    let rec parse1 = parser 
     | [< 'Genlex.Kwd "("; n = parse; 'Genlex.Kwd ")" >] -> n
     | [< 'Genlex.Kwd "eventually"; af = StateFormula.parse >] -> F_temp (TF_eventually af)
     | [< 'Genlex.Kwd "F"; af = StateFormula.parse >] ->  F_temp (TF_eventually af)
     | [< 'Genlex.Kwd "always"; af = StateFormula.parse >] ->  F_temp (TF_always af)
     | [< 'Genlex.Kwd "G"; af = StateFormula.parse >] ->  F_temp (TF_always af)
     | [< 'Genlex.Kwd "not"; n = parse; >] -> F_not(n)
     | [< 'Genlex.Kwd "!"; n = parse; >] -> F_not(n)
    and parse = parser
	| [< n1 = parse1; n2 = parse_remainder n1 >] -> n2

    and parse_remainder n1 = parser
     | [< 'Kwd "or"; n2 = parse >] -> F_or(n1,n2)
     | [< 'Kwd "|"; n2 = parse >] -> F_or(n1,n2)
     | [< 'Kwd "and"; n2 = parse >] -> F_and(n1,n2)
     | [< 'Kwd "&"; n2 = parse >] -> F_and(n1,n2)
     | [< 'Kwd "implies"; n2 = parse >] -> F_implies(n1,n2)
     | [< 'Kwd "=>"; n2 = parse >] -> F_implies(n1,n2)
     | [< >] -> n1

    in parse stream

  let compare (x:t) (y:t) = compare x y
end
 

(* Conversion from weak Muller specification to automata *)

module AFSet = Set.Make(StateFormula)

module NormalizedFormula =
struct 
  type t =
  | AS_and of t * t
  | AS_not of t
  | AS_eventually of StateFormula.t

  let of_formula =
    let rec aux = function
      | Formula.F_or(f,g) -> AS_not (AS_and(aux_not f, aux_not g))
      | F_and(f,g) -> AS_and(aux f,aux g)
      | F_implies(f,g) -> AS_not (AS_and(aux f,aux_not g))
      | F_not(f) -> aux_not f
      | F_temp(TF_always af) -> AS_not (AS_eventually (AF_not af))
      | F_temp(TF_eventually af) -> AS_eventually af
    and aux_not = function
      | Formula.F_or(f,g) -> AS_and(aux_not f, aux_not g)
      | F_and(f,g) -> AS_not (AS_and(aux f,aux g))
      | F_implies(f,g) -> AS_and(aux f,aux_not g)
      | F_not(f) -> aux f
      | F_temp(TF_always af) -> AS_eventually (AF_not af)
      | F_temp(TF_eventually af) -> AS_not (AS_eventually af)
    in aux

  let compare (x:t) (y:t) = compare x y

  let rec atomic_formulas = function
    | AS_and(f,g) -> AFSet.union (atomic_formulas f) (atomic_formulas g)
    | AS_not(f) -> atomic_formulas f
    | AS_eventually(af) -> AFSet.singleton af

  let to_string = 
    let rec aux = function
      | AS_and (a,b) -> "("^aux a^" and "^aux b^")"
      | AS_not a -> "(not "^aux a^")"
      | AS_eventually a -> "(F "^StateFormula.to_string a^")"
    in aux 
end


let formula_to_automaton symbol_size f =
  let aiger = Aiger.empty in
  let f = NormalizedFormula.of_formula f in

  (* there will be one new latch for each atomic formula *)
  let latches = NormalizedFormula.atomic_formulas f in

  (* inputs are the names that appear in the atomic formula *)
  let inputs = AFSet.fold
    (fun x -> StringSet.union (StateFormula.names x)) 
    latches StringSet.empty  
  in

  (* add the inputs to the aiger file *)
  let aiger = StringSet.fold 
    (fun name a -> 
      let rec aux accu i = 
	if i = -1 then accu
	else
	  let na,nv = Aiger.new_var accu in
	  aux (Aiger.add_input na (Aiger.var2lit nv) (name,Some i))
	    (i-1)
      in
      let size = symbol_size name in
      if size = 1 
      then 
	let na,nv = Aiger.new_var a in
	Aiger.add_input na (Aiger.var2lit nv) (name,None)
      else aux a (size - 1)
    ) inputs aiger
  in

  (* add the update function for each latch *)
  let make_update af aiger =
    let bdd = StateFormula.to_bdd (Aiger.name_to_literals aiger) af in
    let aiger,lit = AigerBdd.bdd_to_aiger aiger bdd in
    let aiger,gate = Aiger.new_var aiger in
    let aiger,latch = Aiger.new_var aiger in
    let aiger = Aiger.add_and aiger (Aiger.var2lit gate) (Aiger.aiger_not (Aiger.var2lit latch)) (Aiger.aiger_not lit) in
    Aiger.add_latch aiger (Aiger.var2lit latch) (Aiger.aiger_not (Aiger.var2lit gate)) (StateFormula.to_name af, None)
    
  in 
  let aiger = AFSet.fold make_update latches aiger in
  
  let rec make_winning nf aiger = match nf with
    | NormalizedFormula.AS_and (x,y) ->
      let aiger, litx = make_winning x aiger in 
      let aiger, lity = make_winning y aiger in 
      let aiger,v = Aiger.new_var aiger in
      let lit = Aiger.var2lit v in
      Aiger.add_and aiger lit litx lity, lit
    | AS_not x -> 
      let aiger, lit = make_winning x aiger in 
      aiger, Aiger.aiger_not lit
    | AS_eventually af -> aiger, Aiger.symbol2lit aiger (StateFormula.to_name af,None)

  in 
  let aiger,lit = make_winning f aiger in
  let aiger,v = Aiger.new_var aiger in
  let new_lit = Aiger.var2lit v in
  Aiger.add_latch aiger new_lit lit ("FAIRNESS",None)


module Valuation = 
struct
  module M = Map.Make(StateFormula)

  type t = bool M.t
  let fold = M.fold
  let iter = M.iter
  let map = M.map
  let empty = M.empty
  let find = M.find

  let set valuation form bool =
    M.add form bool valuation

  let to_bdd valuation aiger = 
    fold (fun f b bdd -> 
      let bdd_f = Cudd.ithVar (Aiger.lit2int (Aiger.symbol2lit aiger (StateFormula.to_name f, None))) in
      if b then Cudd.bddAnd bdd bdd_f
      else Cudd.bddAnd bdd (Cudd.bddNot bdd_f)
    ) valuation (Cudd.bddTrue())

  let write outch = 
    iter (fun x b -> 
      Printf.fprintf outch "%s %b\n" (StateFormula.to_string x) b)

  let init formulas =
    AFSet.fold (fun f v -> set v f false) formulas empty

  (* Given a normalized formula and a valuation of the atomic 
     formula inside, tells if the formula is true *)
  let evaluate valuation atomic_formula = 
    let rec aux = function 
      | NormalizedFormula.AS_and(f,g) -> aux f && aux g
      | AS_not(f) -> not (aux f)
      | AS_eventually(af) -> M.find af valuation
    in aux atomic_formula 

end

(*
let winning_region_product aiger controllables uncontrollables formula = 
  let f = NormalizedFormula.of_formula formula in
  let formulas = NormalizedFormula.atomic_formulas f in
  Common.debug (NormalizedFormula.to_string f);
  let init_valuation = Valuation.init formulas in
  let cache = Hashtbl.create (1 lsr AFSet.cardinal formulas) in
  let game = AigerBdd.of_aiger aiger in

  (* next valuation to be evaluated *)
  let successor_valuations valuation = 
    Valuation.fold
      (fun f b x -> 
	if not b then Valuation.set valuation f true :: x else x
      ) valuation [] 
  in
  
  let safe = ref (Cudd.bddFalse ()) in 
  let unsafe = ref (Cudd.bddFalse ()) in

  (* return the winning region and losing region corresponding to each valuation *)
  let rec aux valuation = 
    try Hashtbl.find cache valuation 
    with Not_found ->
      print_endline "there may be a problem in this algorithm with alternations";

      let succ = successor_valuations valuation in
      List.iter aux succ;
      let bdd = Valuation.to_bdd valuation aiger in
      Valuation.write stdout valuation;

      if succ = [] 
      then 
	if Valuation.evaluate valuation f
	then safe := Cudd.bddOr !safe bdd
	else unsafe := Cudd.bddOr !unsafe bdd
      else
	let win = 
	  if Valuation.evaluate valuation f
	  then 
	    let lose = Attractor.trap game controllables uncontrollables !unsafe in
	    (Cudd.bddNot lose)
	  else 
	    let win = Attractor.attractor game controllables uncontrollables !safe
	    in
	    win
	in 
	safe := Cudd.bddOr !safe (Cudd.bddAnd bdd win);
	unsafe := Cudd.bddOr !unsafe (Cudd.bddAnd bdd (Cudd.bddNot win));
	Hashtbl.add cache valuation () ;
  in
  
  let () = aux init_valuation in
  let bdd_init = Valuation.to_bdd init_valuation aiger in
  Cudd.bddRestrict !safe bdd_init, Cudd.bddRestrict !unsafe bdd_init

*)

let winning_region aiger contr uncontr f weak = 
  let auto = formula_to_automaton (Aiger.size_symbol aiger) f in
  
  if !Common.display_debug then 
    (print_endline "Writing automaton to automaton.aag";
     let f = open_out "automaton.aag" in
     Aiger.write auto f; close_out f;);
  
  let product = Aiger.compose aiger auto in
  
  if !Common.display_debug then 
    (print_endline "Writing product game to product.aag";
     let f = open_out "product.aag" in
     Aiger.write product f; close_out f;);
  
  product, Parity.parity2 (AigerBdd.of_aiger product) contr uncontr (AigerBdd.bdd_of_variable (AigerBdd.variable_of_output (Aiger.symbol2lit product ("FAIRNESS",None)))) weak (Attractor.tt())
  

let test aiger_file ?(strong=false) formula = 
  let aiger,contr,uncontr = Attractor.of_aiger aiger_file in
  let f = Formula.parse formula in

  print_endline (Formula.to_string f);
  Cudd.init 100;

  let product,winning = winning_region aiger contr uncontr f strong in
  let initial = AigerBdd.initial_state product in
  let initial_winning = Cudd.bddRestrict (Attractor.latch_configuration winning) initial in

  if Cudd.value initial_winning = 1 
  then print_endline "realizable"
  else if Cudd.value initial_winning = 0
  then print_endline "unrealizable"
  else Cudd.dumpDot ("initial_winning.out") initial_winning


let main = 
  if Filename.basename Sys.argv.(0) = "weakMuller" 
  || Filename.basename Sys.argv.(0) = "weakMuller.byte" 
  then
    if Array.length Sys.argv < 3
    then
      Printf.printf "usage : %s aiger_file formula [?strong]\n" Sys.argv.(0)
    else
      let strongly_winning = if Array.length Sys.argv = 4 then Sys.argv.(3) = "strong" else false in
      let aiger = Aiger.read_from_file Sys.argv.(1) in
      test aiger ~strong:strongly_winning Sys.argv.(2) 

