let debug () = true

type payoff = (WeakMuller.Formula.t * int) list
type player = Coalition.Player.t

module PlayerMap = Map.Make(Coalition.Player)

type payoff_vector = int PlayerMap.t
type payoff_definition = payoff PlayerMap.t

(*
(* the definition should either end with "endrewards" or the end of the file *)
let read_payoff inch =
  let rec loop map nb_line = 
    try 
      let l = input_line inch in
      match Str.split (Str.regexp " *-> *") l with
      | formula_string :: int_string :: [] -> 
	print_endline "parsing weak Muller";
	let formula = WeakMuller.Formula.parse formula_string in 
	let int = int_of_string int_string in
	Printf.printf "payoff %d\n"int;
	loop ((formula,int):: map) (nb_line + 1)

      | "endrewards" :: _ -> map
      | x :: _ -> 
	Printf.printf "unexpected: %S\n" x;
	failwith "syntax error: the syntax for payoff definitions is: <formula> -> <int>"
      | [] -> loop map (nb_line + 1)
    with End_of_file -> map
    | x -> Printf.printf "error in payoff file at line %d of the definition\n" nb_line; map (* raise x*)
  in loop [] 1
  
let read_payoff_definition inch = 
  let rec loop map = 
    try
      let l = input_line inch in
      if l = "" then loop map
      else
	match Str.split (Str.regexp "[ \t]") l with
	| "rewards" :: player_string :: [] -> 
	  let player = int_of_string player_string in
	  let payoff = read_payoff inch in
	  loop (PlayerMap.add player payoff map)
	| _ -> failwith "syntax error: the syntax for payoff definition is: rewards <int>"
    with End_of_file -> map
  in loop PlayerMap.empty

let read_payoff_file file_name = 
  let inch = open_in file_name in
  let def = read_payoff_definition inch in
  close_in inch;
  def

let possible_payoffs payoff_definition =
  PlayerMap.fold
    (fun player payoff accu -> 
      List.fold_left
	(fun list (f,i) -> 
	  (List.map (fun x -> PlayerMap.add player i x) accu) @ list
	) [] payoff
    ) payoff_definition [PlayerMap.empty] 

let payoff_vector_to_string payoff_vector =
  let buf = Buffer.create 100 in
  PlayerMap.fold 
    (fun p i is_first -> 
     if not is_first then Printf.bprintf buf " ; ";
     Printf.bprintf buf "%d:%d" p i; false
    ) payoff_vector true;
  Buffer.contents buf
*)

(*type comparison = Smaller | SmallerEq | Equal | GreaterEq | Greater

let comparison_to_string = function
  | Smaller -> "<" | SmallerEq -> "<=" | Equal -> "="
  | GreaterEq -> ">=" | Greater -> ">"

let comparison_of_string = function
  | "<" -> Smaller | "<=" -> SmallerEq | "=" -> Equal
  | ">=" -> GreaterEq | ">" -> Greater | _ -> raise Not_found

let comparison_function = function
  | Smaller -> (<) | SmallerEq -> (<=) | Equal -> (=)
  | GreaterEq -> (>=) | Greater -> (>)
*)
type payoff_formula = 
| PF_not of payoff_formula
| PF_and of payoff_formula * payoff_formula
| PF_or of payoff_formula * payoff_formula
(*| PF_compare of player * comparison * int*)
(*| PF_compare of string * comparison * int*)
| PF_eventually of StateFormula.t
| PF_always of StateFormula.t
(*| PF_state of StateFormula.t*)

type formula = (Coalition.t * payoff_formula) list


let rec payoff_formula_to_string = function
  | PF_not pf -> Printf.sprintf "(not %s)" (payoff_formula_to_string pf)
  | PF_and (pf1,pf2) -> Printf.sprintf "(%s and %s)" (payoff_formula_to_string pf1) (payoff_formula_to_string pf2)
  (*| PF_compare (p,c,i) -> Printf.sprintf "payoff(%d) %s %d" p (comparison_to_string c) i*)
(*  | PF_compare (p,c,i) -> Printf.sprintf "%s %s %d" p (comparison_to_string c) i*)
  | PF_or (f1,f2) -> Printf.sprintf "(%s or %s)" (payoff_formula_to_string f1) (payoff_formula_to_string f2)
  | PF_always sf -> Printf.sprintf "G { %s }" (StateFormula.to_string sf)
  | PF_eventually sf -> Printf.sprintf "F { %s }" (StateFormula.to_string sf)

let formula_to_string f = 
  let buffer = Buffer.create 80 in
  List.iter (fun (c,pf) ->
    Buffer.add_string buffer "([";
    Buffer.add_string buffer (Coalition.to_string c);
    Buffer.add_string buffer "] ";
    Buffer.add_string buffer (payoff_formula_to_string pf);
    Buffer.add_string buffer ")\n";
  ) f;
  Buffer.contents buffer


(*let parse_compare = 
  let rec parse = parser
(*  | [< 'Genlex.Kwd "payoff"; 'Genlex.Kwd "("; 'Genlex.Int p; 'Genlex.Kwd ")"; 'Genlex.Kwd c; 'Genlex.Int i >] -> PF_compare (p,comparison_of_string c,i)*)
(*  | [< 'Genlex.Ident p; 'Genlex.Kwd c; 'Genlex.Int i >] -> PF_compare (p,comparison_of_string c,i)*)
  | [< >] -> failwith "parsing failed: waiting \"{state formula} \" "
  in parse*)

let parse_payoff_formula =   
  let rec parse_remainder r = parser
  | [< 'Genlex.Kwd "and"; s=parse >] -> PF_and(r,s)
  | [< >] -> r
  and parse = parser 
      | [< 'Genlex.Kwd "("; n = parse; 'Genlex.Kwd ")" >] -> n
      | [< 'Genlex.Kwd "not"; n = parse; >] -> PF_not n
      | [< 'Genlex.Kwd "{"; c = StateFormula.parse; 'Genlex.Kwd "}"; pf = parse_remainder (PF_always c) >] -> pf
      | [< 'Genlex.Kwd "G";'Genlex.Kwd "{"; c = StateFormula.parse; 'Genlex.Kwd "}"; pf = parse_remainder (PF_always c) >] -> pf
      | [< 'Genlex.Kwd "F"; 'Genlex.Kwd "{"; c = StateFormula.parse; 'Genlex.Kwd "}"; pf = parse_remainder (PF_eventually c) >] -> pf
      | [< >] -> failwith "parse formula failed: waiting \"(\"  \"not\" \"G\" or \"F {\""
  in parse

let parse_formula string =
  let lexer = Genlex.make_lexer ["+";"-";"or";"not";"and";"implies";"="; "<="; "<"; ">"; ">=";"("; ")"; "deviation"; "payoff";";";"[";"]";"{";"}";"F";"G"]  in
  let stream = lexer (Stream.of_string string) in
  
  let rec parse_players = parser 
  | [< 'Genlex.Int p ; c = parse_players_remaining (Coalition.singleton p)>] -> c
  | [< >] -> Coalition.empty
(*failwith "parse formula failed: waiting player identifier"*)
  and parse_players_remaining accu = parser
      | [< 'Genlex.Kwd ";"; 'Genlex.Int p ; s = parse_players_remaining (Coalition.add p accu) >] -> s
      | [< >] -> accu
  in
  let rec parse = parser 
  | [< 'Genlex.Kwd "["; p = parse_players; 'Genlex.Kwd "]"; pf = parse_payoff_formula; n2 = parse_remainder [p,pf] >] -> n2
  | [< 'Genlex.Kwd "("; p = parse; 'Genlex.Kwd ")"; n2 = parse_remainder p >] -> n2
  | [< >] -> failwith "parse formula failed: waiting \"[players] payoff formula\""

  and parse_remainder accu = parser
      | [< 'Genlex.Kwd "and"; n2 = parse >] -> List.rev_append n2 accu
      | [< >] -> accu
       
  in 
  try parse stream 
  with x -> 
    Printf.eprintf "next token:%s\n" (Common.next_token stream);
    raise x


exception TrivialFormula of bool
(*
let compare_to_atomic (*payoff_definition*) (p,comparison,int) =
  match comparison with
  | Smaller -> 
    StateFormula.of_smaller_equal p (int-1)
  | SmallerEq -> 
    WeakMuller.AtomicFormula.of_smaller_equal p int
  | Equal ->
    let f1 = WeakMuller.AtomicFormula.of_greater_equal p int in
    let f2 = WeakMuller.AtomicFormula.of_smaller_equal p int in
    WeakMuller.AtomicFormula.conjunction f1 f2
  | GreaterEq -> 
    WeakMuller.AtomicFormula.of_greater_equal p int
  | Greater -> 
    WeakMuller.AtomicFormula.of_greater_equal p (int+1)
*)
  (* select payoffs that are greater than [i] *)
(*  let greater_eq i =
    let rec aux accu = function
      | [] -> accu
      | (fj,j) :: s -> 
	if j >= i then aux (fj :: accu) s else aux accu s
    in aux []
  in
  
  let at_least_eq i = 
    let rec aux accu = function
      | [] -> accu
      | (fj,j) :: s -> 
	if j = i then aux (fj :: accu) s else aux accu s
    in aux []
  in

  let list_to_formula = function
    | [] -> raise (TrivialFormula false)
    | hd :: tl -> 
      List.fold_left
	(fun f fj -> WeakMuller.Formula.F_or (f,fj)) hd tl
  in
   let definition = PlayerMap.find player payoff_definition in*)
 (* match comparison with
  | Smaller -> 
     (try WeakMuller.Formula.F_not (list_to_formula (greater_eq int definition))
     with TrivialFormula b -> raise (TrivialFormula (not b)))
  | SmallerEq -> 
    (try WeakMuller.Formula.F_not (list_to_formula (greater_eq (int+1) definition))
     with TrivialFormula b -> raise (TrivialFormula (not b)))
  | Equal -> 
    let f1 = list_to_formula (at_least_eq int definition) in
    (try
       let f2 = WeakMuller.Formula.F_not (list_to_formula (greater_eq (int+1) definition)) in
       WeakMuller.Formula.F_and (f1, f2)
     with TrivialFormula false -> f1 | TrivialFormula true -> raise ( TrivialFormula false))
       
  | GreaterEq -> list_to_formula (greater_eq int definition)
  | Greater -> list_to_formula (greater_eq (int+1) definition)*)




let payoff_formula_to_weak_Muller (*payoff_definition*) payoff_formula =
  Common.debug "converting formula:";
  Common.debug (payoff_formula_to_string payoff_formula);
  let rec aux = function 
    | PF_not pf -> 
      (try WeakMuller.Formula.F_not (aux pf)
       with TrivialFormula b -> raise (TrivialFormula (not b)))

    | PF_and (pf1,pf2) -> 
      (try let r1 = aux pf1 in
	   try let r2 = aux pf2 in
	       WeakMuller.Formula.F_and (r1,r2)
	   with TrivialFormula true -> r1
       with TrivialFormula true -> aux pf2)

    | PF_or (pf1,pf2) -> 
      (try let r1 = aux pf1 in
	   try let r2 = aux pf2 in
	       WeakMuller.Formula.F_or (r1,r2)
	   with TrivialFormula false -> r1
       with TrivialFormula false -> aux pf2)

    | PF_always sf -> WeakMuller.Formula.F_temp (TF_always sf)
    | PF_eventually sf -> WeakMuller.Formula.F_temp (TF_eventually sf)
    (*| PF_compare (p,c,i) -> 
      compare_to_atomic (p,c,i)*)
	
  in aux payoff_formula


let formula_to_weak_Muller (*payoff_definition*) players formula = 
  (* let players = PlayerMap.fold (fun p i c -> Coalition.add p c) payoff_definition Coalition.empty in*)
  (* formula that tells that [x] has not deviated *)
  let not_deviator x = WeakMuller.Formula.F_temp (WeakMuller.Formula.TF_always (StateFormula.AF_equal ((Translation.deviator x),0))) in

  (* convert one formula of the list *)
  let rec aux (c,pf) = 
    let elements = Coalition.elements (Coalition.diff players c) in
    match elements with 
    | [] -> payoff_formula_to_weak_Muller (*payoff_definition*) pf
    | hd :: tl ->
      (* formula that states that the set of deviators is included in the coaltion c*)
      let inclusion = (* always deviator hd = 0 /\ always deviator tl = 0 *)
	List.fold_left (fun x p -> WeakMuller.Formula.F_and (x,not_deviator p)) (not_deviator hd) tl
      in
      WeakMuller.Formula.F_implies (inclusion,payoff_formula_to_weak_Muller (*payoff_definition*) pf)
  in 

  let conjunction = function 
    | [] -> raise (TrivialFormula true)
    | hd :: tl ->
      List.fold_left (fun accu x -> WeakMuller.Formula.F_and (accu,x)) hd tl
  in

  let formula_list = List.fold_left (fun accu x ->
    try aux x :: accu with TrivialFormula true -> accu
  ) [] formula 
  in conjunction formula_list


let equilibria_region multi_game players (*payoff_definition*) formula = 
  let game,controllables,uncontrollables = Translation.deviator_game multi_game in
  print_endline "some latches appear two time in the product";
  
(*  if !Common.display_debug then *)
    (print_endline "Writing translated game to deviator.aag";
     let f = open_out "deviator.aag" in
     Aiger.write game f; close_out f;);

  let condition = formula_to_weak_Muller players formula (*payoff_definition formula*) in
  Common.infos (WeakMuller.Formula.to_short_string condition);
  Common.debug "computing winning region";
    (* Warning : we have to make sure the players play in 
       the right order. In weakMuller the environment moves
       first so we invert controllables and uncontrollables,  
       negate the condition and invert winning and losing.
    *)
  WeakMuller.winning_region game uncontrollables controllables (WeakMuller.Formula.F_not condition) true 
    



let test multi_game formula = 
  (*Cudd.init 100;*)
  let game,controllables,uncontrollables = Translation.deviator_game multi_game in
  print_endline "Writing translated game to deviator.aag";
  let f = open_out "deviator.aag" in
  Aiger.write game f; 
  close_out f;

  let condition = formula_to_weak_Muller (Coalition.of_aiger multi_game.Translation.aiger) formula in
  let cond_string = WeakMuller.Formula.to_string condition in
  print_endline cond_string;
  (* WeakMuller.test game ~strong:true cond_string; *)
  WeakMuller.test game ~strong:false cond_string

  (*
  let payoff_definition = read_payoff_file Sys.argv.(2) in
  print_endline "payoff read";*)

(*  
  let game,winning = equilibria_region multi_game (players multi_game.aiger) (*payoff_definition*) formula in
  if debug () then print_endline "computing initial state";

  let initial = AigerBdd.initial_state game in
  let initial_winning = Cudd.bddRestrict winning initial in
    
  if Cudd.value initial_winning = 1 
  then print_endline "realizable"
  else if Cudd.value initial_winning = 0
  then print_endline "unrealizable"
  else Cudd.dumpDot ("initial_winning.out") initial_winning
*)      


  
let main = 
  if Filename.basename Sys.argv.(0) = "weakMullerEquilibria" 
     || Filename.basename Sys.argv.(0) = "weakMullerEquilibria.byte" 
  then 
    if Array.length Sys.argv < 3
    then Printf.printf "usage : %s aiger_file formula\n" Sys.argv.(0)
    else   
      let formula = parse_formula Sys.argv.(2) in
      print_string "formula to check : ";
      print_endline (formula_to_string formula);
      let multi_game = Translation.read_from_file Sys.argv.(1) in
      Printf.printf "multiplayer game read from %s\n" Sys.argv.(1);
      test multi_game formula
    
